import pytest
from nnsvs.acoustic_models import (
    BiLSTMMDNNonAttentiveDecoder,
    BiLSTMNonAttentiveDecoder,
    BiLSTMResF0NonAttentiveDecoder,
    MDNMultistreamSeparateF0MelModel,
    MDNNonAttentiveDecoder,
    MDNResF0NonAttentiveDecoder,
    MultistreamSeparateF0MelModel,
    MultistreamSeparateF0ParametricModel,
    NonAttentiveDecoder,
    NPSSMDNMultistreamParametricModel,
    NPSSMultistreamParametricModel,
    ResF0Conv1dResnet,
    ResF0Conv1dResnetMDN,
    ResF0NonAttentiveDecoder,
    ResF0TransformerEncoder,
    ResF0VariancePredictor,
    ResSkipF0FFConvLSTM,
)
from nnsvs.base import PredictionType
from nnsvs.model import FFN, MDN, FFConvLSTM, LSTMEncoder, TransformerEncoder

from .util import _test_model_impl


def test_deprecated_imports():
    from nnsvs.acoustic_models import LSTMEncoder  # noqa: F401


@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
def test_resf0_conv1d_resnet(num_gaussians):
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 200,
        "num_layers": 2,
        "num_gaussians": num_gaussians,
        "dim_wise": True,
        "init_type": "none",
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = ResF0Conv1dResnet(**{**params, "use_mdn": False})
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    model = ResF0Conv1dResnet(**{**params, "use_mdn": True})
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    # Deprecated
    model = ResF0Conv1dResnetMDN(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
def test_reskipf0_ff_conv_lstm(num_gaussians):
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 8,
        "conv_hidden_dim": 8,
        "lstm_hidden_dim": 8,
        "dropout": 0.1,
        "num_lstm_layers": 2,
        "bidirectional": True,
        "out_dim": 200,
        "num_gaussians": num_gaussians,
        "dim_wise": True,
        "init_type": "none",
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = ResSkipF0FFConvLSTM(**{**params, "use_mdn": False})
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    model = ResSkipF0FFConvLSTM(**{**params, "use_mdn": True})
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
def test_resf0_variance_predictor(num_gaussians):
    params = {
        "in_dim": 300,
        "out_dim": 200,
        "num_layers": 2,
        "hidden_dim": 8,
        "kernel_size": 5,
        "dropout": 0.5,
        "num_gaussians": num_gaussians,
        "dim_wise": True,
        "init_type": "none",
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = ResF0VariancePredictor(**{**params, "use_mdn": False})
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    model = ResF0VariancePredictor(**{**params, "use_mdn": True})
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
@pytest.mark.parametrize("use_mdn_lf0", [False, True])
@pytest.mark.parametrize("vuv_model_lf0_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_mel_conditioning", [False, True])
def test_hybrid_multistream_mel_model_vuv_pred_from_mel(
    reduction_factor,
    num_gaussians,
    use_mdn_lf0,
    vuv_model_lf0_conditioning,
    vuv_model_mel_conditioning,
):
    vuv_in_dim = 300
    if vuv_model_lf0_conditioning:
        vuv_in_dim += 1
    if vuv_model_mel_conditioning:
        vuv_in_dim += 80
    params = {
        "in_dim": 300,
        "out_dim": 82,
        "stream_sizes": [80, 1, 1],
        "reduction_factor": reduction_factor,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
            use_mdn=use_mdn_lf0,
            num_gaussians=num_gaussians,
        ),
        # Decoders
        "mel_model": MDN(
            in_dim=301,
            hidden_dim=5,
            out_dim=80,
            dim_wise=True,
            num_gaussians=num_gaussians,
        ),
        "vuv_model": FFN(in_dim=vuv_in_dim, hidden_dim=5, out_dim=1),
        "vuv_model_lf0_conditioning": vuv_model_lf0_conditioning,
        "vuv_model_mel_conditioning": vuv_model_mel_conditioning,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MDNMultistreamSeparateF0MelModel(**params)
    assert model.prediction_type() == PredictionType.MULTISTREAM_HYBRID
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
def test_multistream_mel_model(reduction_factor):
    params = {
        "in_dim": 300,
        "out_dim": 82,
        "stream_sizes": [80, 1, 1],
        "reduction_factor": reduction_factor,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
        ),
        # Encoder
        "encoder": LSTMEncoder(
            in_dim=300,
            hidden_dim=2,
            out_dim=4,
            num_layers=2,
            dropout=0.5,
            init_type="none",
        ),
        # Decoders
        "mel_model": FFN(in_dim=6, hidden_dim=5, out_dim=80),
        "vuv_model": FFN(in_dim=6, hidden_dim=5, out_dim=1),
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MultistreamSeparateF0MelModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
@pytest.mark.parametrize("vuv_model_bap0_conditioning", [False, True])
@pytest.mark.parametrize("use_mdn_lf0", [False, True])
@pytest.mark.parametrize("vuv_model_bap_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_lf0_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_mgc_conditioning", [False, True])
def test_npss_mdn_multistream_parametric_model(
    num_gaussians,
    vuv_model_bap0_conditioning,
    use_mdn_lf0,
    vuv_model_bap_conditioning,
    vuv_model_lf0_conditioning,
    vuv_model_mgc_conditioning,
):
    vuv_in_dim = 300
    if vuv_model_lf0_conditioning:
        vuv_in_dim += 1
    if vuv_model_mgc_conditioning:
        vuv_in_dim += 60
    if vuv_model_bap_conditioning:
        if vuv_model_bap0_conditioning:
            vuv_in_dim += 1
        else:
            vuv_in_dim += 5
    params = {
        "in_dim": 300,
        "out_dim": 67,
        "stream_sizes": [60, 1, 1, 5],
        "reduction_factor": 1,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=1,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
            use_mdn=use_mdn_lf0,
            num_gaussians=num_gaussians,
        ),
        # Decoders
        "mgc_model": MDN(
            in_dim=301,
            hidden_dim=5,
            out_dim=60,
            dim_wise=True,
            num_gaussians=num_gaussians,
        ),
        "bap_model": MDN(
            in_dim=301,
            hidden_dim=5,
            out_dim=5,
            dim_wise=True,
            num_gaussians=num_gaussians,
        ),
        "vuv_model": FFN(
            in_dim=vuv_in_dim,
            hidden_dim=5,
            out_dim=1,
        ),
        "vuv_model_bap_conditioning": vuv_model_bap_conditioning,
        "vuv_model_bap0_conditioning": vuv_model_bap0_conditioning,
        "vuv_model_lf0_conditioning": vuv_model_lf0_conditioning,
        "vuv_model_mgc_conditioning": vuv_model_mgc_conditioning,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = NPSSMDNMultistreamParametricModel(**params)
    assert model.prediction_type() == PredictionType.MULTISTREAM_HYBRID
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("vuv_model_bap0_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_bap_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_lf0_conditioning", [False, True])
@pytest.mark.parametrize("vuv_model_mgc_conditioning", [False, True])
def test_nonmdn_npss_multistream_parametric_model(
    vuv_model_bap0_conditioning,
    vuv_model_bap_conditioning,
    vuv_model_lf0_conditioning,
    vuv_model_mgc_conditioning,
):
    vuv_in_dim = 300
    if vuv_model_lf0_conditioning:
        vuv_in_dim += 1
    if vuv_model_mgc_conditioning:
        vuv_in_dim += 60
    if vuv_model_bap_conditioning:
        if vuv_model_bap0_conditioning:
            vuv_in_dim += 1
        else:
            vuv_in_dim += 5
    params = {
        "in_dim": 300,
        "out_dim": 67,
        "stream_sizes": [60, 1, 1, 5],
        "reduction_factor": 1,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=1,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
            use_mdn=False,
        ),
        # Decoders
        "mgc_model": FFN(
            in_dim=301,
            hidden_dim=5,
            out_dim=60,
        ),
        "bap_model": FFN(
            in_dim=301,
            hidden_dim=5,
            out_dim=5,
        ),
        "vuv_model": FFN(
            in_dim=vuv_in_dim,
            hidden_dim=5,
            out_dim=1,
        ),
        "vuv_model_bap_conditioning": vuv_model_bap_conditioning,
        "vuv_model_bap0_conditioning": vuv_model_bap0_conditioning,
        "vuv_model_lf0_conditioning": vuv_model_lf0_conditioning,
        "vuv_model_mgc_conditioning": vuv_model_mgc_conditioning,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = NPSSMultistreamParametricModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("num_gaussians", [1, 2, 4])
def test_bilstm_mdn_tacotron_decoder(
    reduction_factor, downsample_by_conv, num_gaussians
):
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 4,
        "conv_hidden_dim": 4,
        "lstm_hidden_dim": 4,
        "num_lstm_layers": 1,
        "out_dim": 60,
        "decoder_hidden_dim": 4,
        "decoder_layers": 1,
        "prenet_layers": 0,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        "num_gaussians": num_gaussians,
    }
    model = BiLSTMMDNNonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("use_mdn", [False, True])
def test_ff_conv_lstm(use_mdn):
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 8,
        "conv_hidden_dim": 8,
        "lstm_hidden_dim": 8,
        "dropout": 0.1,
        "num_lstm_layers": 2,
        "bidirectional": True,
        "out_dim": 200,
        "init_type": "none",
        "use_mdn": use_mdn,
        "num_gaussians": 2,
        "dim_wise": True,
    }
    model = FFConvLSTM(**params)
    if use_mdn:
        assert model.prediction_type() == PredictionType.PROBABILISTIC
    else:
        assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1])
def test_resf0_tacotron_decoder_mdn(
    reduction_factor, downsample_by_conv, prenet_layers
):
    params = {
        "in_dim": 300,
        "out_dim": 80,
        "hidden_dim": 4,
        "layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        "num_gaussians": 4,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 0,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MDNResF0NonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1, 2])
@pytest.mark.parametrize("sampling_mode", ["mean", "random"])
def test_tacotron_decoder_mdn(
    reduction_factor, downsample_by_conv, prenet_layers, sampling_mode
):
    params = {
        "in_dim": 300,
        "out_dim": 206,
        "hidden_dim": 4,
        "layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        "num_gaussians": 8,
        "sampling_mode": sampling_mode,
        "initial_value": 0.0,
    }
    model = MDNNonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("num_heads", [1, 2, 4])
@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
def test_transformer_encoder(num_heads, reduction_factor, downsample_by_conv):
    params = {
        "in_dim": 300,
        "out_dim": 206,
        "hidden_dim": 4,
        "attention_dim": 4 * num_heads,
        "num_heads": num_heads,
        "num_layers": 2,
        "kernel_size": 3,
        "dropout": 0.1,
        "reduction_factor": reduction_factor,
        "init_type": "none",
        "downsample_by_conv": downsample_by_conv,
    }
    model = TransformerEncoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1, 2])
def test_tacotron_decoder(reduction_factor, downsample_by_conv, prenet_layers):
    params = {
        "in_dim": 300,
        "out_dim": 206,
        "hidden_dim": 4,
        "layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        "initial_value": 0.0,
    }
    model = NonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1])
def test_bilstm_tacotron_decoder(reduction_factor, downsample_by_conv, prenet_layers):
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 4,
        "conv_hidden_dim": 4,
        "lstm_hidden_dim": 4,
        "num_lstm_layers": 1,
        "out_dim": 60,
        "decoder_hidden_dim": 4,
        "decoder_layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
    }
    model = BiLSTMNonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1])
def test_resf0_tacotron_decoder(reduction_factor, downsample_by_conv, prenet_layers):
    params = {
        "in_dim": 300,
        "out_dim": 3,
        "hidden_dim": 4,
        "layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 0,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = ResF0NonAttentiveDecoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
@pytest.mark.parametrize("prenet_layers", [0, 1])
@pytest.mark.parametrize("use_mdn", [False, True])
def test_bilstm_resf0_tacotron_decoder(
    reduction_factor, downsample_by_conv, prenet_layers, use_mdn
):
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 4,
        "conv_hidden_dim": 4,
        "lstm_hidden_dim": 4,
        "num_lstm_layers": 1,
        "out_dim": 3,
        "decoder_hidden_dim": 4,
        "decoder_layers": 1,
        "prenet_layers": prenet_layers,
        "prenet_hidden_dim": 4,
        "prenet_dropout": 0.5,
        "zoneout": 0.1,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 0,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
        "use_mdn": use_mdn,
    }
    model = BiLSTMResF0NonAttentiveDecoder(**params)
    if use_mdn:
        assert model.prediction_type() == PredictionType.PROBABILISTIC
    else:
        assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("num_heads", [1, 2, 4])
@pytest.mark.parametrize("reduction_factor", [1, 2])
def test_resf0_transformer_encoder(num_heads, reduction_factor):
    params = {
        "in_dim": 300,
        "out_dim": 206,
        "hidden_dim": 4,
        "attention_dim": 4 * num_heads,
        "num_heads": num_heads,
        "num_layers": 2,
        "kernel_size": 3,
        "dropout": 0.1,
        "reduction_factor": reduction_factor,
        "init_type": "none",
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = ResF0TransformerEncoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_lstm_encoder():
    params = {
        "in_dim": 300,
        "out_dim": 206,
        "hidden_dim": 4,
        "num_layers": 2,
        "dropout": 0.1,
        "init_type": "none",
    }
    model = LSTMEncoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
def test_nonar_multistream_parametric_model(reduction_factor):
    params = {
        "in_dim": 300,
        "out_dim": 199,
        "stream_sizes": [180, 3, 1, 15],
        "reduction_factor": reduction_factor,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=3,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
        ),
        # Encoder
        "encoder": LSTMEncoder(
            in_dim=300,
            hidden_dim=2,
            out_dim=4,
            num_layers=2,
            dropout=0.5,
            init_type="none",
        ),
        # Decoders
        "mgc_model": FFN(in_dim=8, hidden_dim=5, out_dim=180),
        "vuv_model": FFN(in_dim=8, hidden_dim=5, out_dim=1),
        "bap_model": FFN(in_dim=8, hidden_dim=5, out_dim=15),
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MultistreamSeparateF0ParametricModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
def test_ar_multistream_parametric_model(reduction_factor):
    params = {
        "in_dim": 300,
        "out_dim": 199,
        "stream_sizes": [180, 3, 1, 15],
        "reduction_factor": reduction_factor,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=3,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
        ),
        # Encoder
        "encoder": LSTMEncoder(
            in_dim=300,
            hidden_dim=2,
            out_dim=4,
            num_layers=2,
            dropout=0.5,
            init_type="none",
        ),
        # Decoders
        "mgc_model": NonAttentiveDecoder(
            in_dim=8,
            hidden_dim=5,
            out_dim=180,
            layers=2,
            reduction_factor=reduction_factor,
        ),
        "vuv_model": NonAttentiveDecoder(
            in_dim=8,
            hidden_dim=5,
            out_dim=1,
            layers=1,
            reduction_factor=reduction_factor,
        ),
        "bap_model": NonAttentiveDecoder(
            in_dim=8,
            hidden_dim=5,
            out_dim=15,
            layers=1,
            reduction_factor=reduction_factor,
        ),
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MultistreamSeparateF0ParametricModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


@pytest.mark.parametrize("reduction_factor", [1, 2])
@pytest.mark.parametrize("postnet_layers", [0, 1])
def test_ar_multistream_parametric_model_with_postnet(reduction_factor, postnet_layers):
    params = {
        "in_dim": 300,
        "out_dim": 67,
        "stream_sizes": [60, 1, 1, 5],
        "reduction_factor": reduction_factor,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
        ),
        # Encoder
        "encoder": LSTMEncoder(
            in_dim=300,
            hidden_dim=2,
            out_dim=4,
            num_layers=2,
            dropout=0.5,
            init_type="none",
        ),
        # Decoders
        "mgc_model": BiLSTMNonAttentiveDecoder(
            in_dim=6,
            out_dim=60,
            ff_hidden_dim=8,
            conv_hidden_dim=8,
            lstm_hidden_dim=8,
            num_lstm_layers=1,
            decoder_layers=1,
            decoder_hidden_dim=8,
            prenet_layers=1,
            prenet_hidden_dim=8,
            postnet_layers=postnet_layers,
            postnet_channels=8,
            postnet_kernel_size=3,
            reduction_factor=reduction_factor,
        ),
        "vuv_model": BiLSTMNonAttentiveDecoder(
            in_dim=6,
            out_dim=1,
            ff_hidden_dim=8,
            conv_hidden_dim=8,
            lstm_hidden_dim=8,
            num_lstm_layers=1,
            decoder_layers=1,
            decoder_hidden_dim=8,
            prenet_layers=1,
            prenet_hidden_dim=8,
            postnet_layers=postnet_layers,
            postnet_channels=8,
            postnet_kernel_size=3,
            reduction_factor=reduction_factor,
        ),
        "bap_model": BiLSTMNonAttentiveDecoder(
            in_dim=6,
            out_dim=5,
            ff_hidden_dim=8,
            conv_hidden_dim=8,
            lstm_hidden_dim=8,
            num_lstm_layers=1,
            decoder_layers=1,
            decoder_hidden_dim=8,
            prenet_layers=1,
            prenet_hidden_dim=8,
            postnet_layers=postnet_layers,
            postnet_channels=8,
            postnet_kernel_size=3,
            reduction_factor=reduction_factor,
        ),
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 60,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MultistreamSeparateF0ParametricModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_nonar_multistream_parametric_model_no_encoder():
    params = {
        "in_dim": 300,
        "out_dim": 67,
        "stream_sizes": [60, 1, 1, 5],
        "reduction_factor": 1,
        # Separate f0 model
        "lf0_model": ResF0Conv1dResnet(
            in_dim=300,
            hidden_dim=5,
            out_dim=1,
            num_layers=1,
            in_lf0_idx=-1,
            out_lf0_idx=0,
        ),
        # No encoder
        "encoder": None,
        # Decoders
        "mgc_model": FFN(in_dim=300, hidden_dim=5, out_dim=60),
        "vuv_model": FFN(in_dim=300, hidden_dim=5, out_dim=1),
        "bap_model": FFN(in_dim=300, hidden_dim=5, out_dim=5),
        # dummy
        "in_lf0_idx": 0,
        "in_lf0_min": 5.3936276,
        "in_lf0_max": 6.491111,
        "out_lf0_idx": 180,
        "out_lf0_mean": 5.953093881972361,
        "out_lf0_scale": 0.23435173188961034,
    }
    model = MultistreamSeparateF0ParametricModel(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    assert not model.is_autoregressive()
    _test_model_impl(model, params["in_dim"], params["out_dim"])



from pathlib import Path

import hydra
import nnsvs.bin.train
import nnsvs.bin.train_acoustic
import nnsvs.bin.train_postfilter
import pytest
import torch
from nnsvs.util import init_seed
from omegaconf import OmegaConf

from .util import _test_model_impl

RECIPE_DIR = Path(__file__).parent.parent / "recipes"


def _test_postfilter_impl(model, model_config):
    B = 4
    T = 100
    init_seed(B * T)

    in_dim = sum(model_config.stream_sizes)
    x = torch.rand(B, T, in_dim)
    lengths = torch.Tensor([T] * B).long()

    # warmup forward pass
    with torch.no_grad():
        y = model(x, lengths)
        y_inf = model.inference(x, lengths)

    assert x.shape == y.shape
    assert y_inf.shape == y.shape


@pytest.mark.parametrize(
    "model_config",
    (Path(nnsvs.bin.train.__file__).parent / "conf" / "train" / "model").glob("*.yaml"),
)
def test_model_config(model_config):
    model_config = OmegaConf.load(model_config)
    model = hydra.utils.instantiate(model_config.netG)
    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config",
    (
        Path(nnsvs.bin.train_acoustic.__file__).parent
        / "conf"
        / "train_acoustic"
        / "model"
    ).glob("*.yaml"),
)
def test_acoustic_model_config(model_config):
    print(model_config)
    model_config = OmegaConf.load(model_config)

    # Dummy
    model_config.netG.in_lf0_idx = 10
    model_config.netG.in_lf0_min = 5.3936276
    model_config.netG.in_lf0_max = 6.491111
    model_config.netG.out_lf0_idx = 60
    model_config.netG.out_lf0_mean = 5.953093881972361
    model_config.netG.out_lf0_scale = 0.23435173188961034

    model = hydra.utils.instantiate(model_config.netG)

    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config",
    (
        Path(nnsvs.bin.train_postfilter.__file__).parent
        / "conf"
        / "train_postfilter"
        / "model"
    ).glob("*.yaml"),
)
def test_postfilter_model_config(model_config):
    model_config = OmegaConf.load(model_config)
    if "stream_sizes" in model_config.netG:
        model_config.netG.stream_sizes = model_config.stream_sizes
    # Post-filter config should have netD
    hydra.utils.instantiate(model_config.netD)
    model = hydra.utils.instantiate(model_config.netG)
    _test_postfilter_impl(model, model_config)


@pytest.mark.parametrize(
    "model_config", RECIPE_DIR.glob("**/_common/conf/**/train/timelag/model/*.yaml")
)
def test_timelag_model_config_recipes(model_config):
    model_config = OmegaConf.load(model_config)
    model = hydra.utils.instantiate(model_config.netG)
    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config", RECIPE_DIR.glob("**/_common/conf/**/train/duration/model/*.yaml")
)
def test_duration_model_config_recipes(model_config):
    model_config = OmegaConf.load(model_config)
    model = hydra.utils.instantiate(model_config.netG)
    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config", RECIPE_DIR.glob("**/_common/conf/**/train_acoustic/model/*.yaml")
)
def test_train_acoustic_model_config_recipes(model_config):
    print(model_config)
    model_config = OmegaConf.load(model_config)

    # Dummy
    model_config.netG.in_lf0_idx = 10
    model_config.netG.in_lf0_min = 5.3936276
    model_config.netG.in_lf0_max = 6.491111
    model_config.netG.out_lf0_idx = 60
    model_config.netG.out_lf0_mean = 5.953093881972361
    model_config.netG.out_lf0_scale = 0.23435173188961034

    # NOTE: phoneme embedding impl requires careful specifications of
    # in_ph_start_idx and in_ph_end_idx, which is diffifult to set property
    # without hed files. So we skip phoneme embedding impl for now.
    if "embed_dim" in model_config.netG:
        model_config.netG.embed_dim = None

    keys = ["lf0_model", "mgc_model", "bap_model", "vuv_model", "encoder", "mel_model"]
    for key in keys:
        if key in model_config.netG and "embed_dim" in model_config.netG[key]:
            model_config.netG[key].embed_dim = None

        # For GaussianDiffusion that includes encoder in its implementation
        if (
            key in model_config.netG
            and "encoder" in model_config.netG[key]
            and "embed_dim" in model_config.netG[key]["encoder"]
        ):
            model_config.netG[key]["encoder"].embed_dim = None

    model = hydra.utils.instantiate(model_config.netG)
    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config",
    RECIPE_DIR.glob("**/icassp2023-24k-world/conf/train_acoustic/model/*.yaml"),
)
def test_train_acoustic_model_config_recipes_icassp2023(model_config):
    model_config = OmegaConf.load(model_config)

    # Dummy
    model_config.netG.in_lf0_idx = 10
    model_config.netG.in_lf0_min = 5.3936276
    model_config.netG.in_lf0_max = 6.491111
    model_config.netG.out_lf0_idx = 60
    model_config.netG.out_lf0_mean = 5.953093881972361
    model_config.netG.out_lf0_scale = 0.23435173188961034

    # NOTE: phoneme embedding impl requires careful specifications of
    # in_ph_start_idx and in_ph_end_idx, which is diffifult to set property
    # without hed files. So we skip phoneme embedding impl for now.
    if "embed_dim" in model_config.netG:
        model_config.netG.embed_dim = None

    keys = ["lf0_model", "mgc_model", "bap_model", "vuv_model", "encoder", "mel_model"]
    for key in keys:
        if key in model_config.netG and "embed_dim" in model_config.netG[key]:
            model_config.netG[key].embed_dim = None

    model = hydra.utils.instantiate(model_config.netG)
    _test_model_impl(model, model_config.netG.in_dim, model_config.netG.out_dim)


@pytest.mark.parametrize(
    "model_config", RECIPE_DIR.glob("**/_common/conf/**/train_postfilter/model/*.yaml")
)
def test_postfilter_config_recipes(model_config):
    model_config = OmegaConf.load(model_config)
    if "stream_sizes" in model_config.netG:
        model_config.netG.stream_sizes = model_config.stream_sizes
    # Post-filter config should have netD
    hydra.utils.instantiate(model_config.netD)
    model = hydra.utils.instantiate(model_config.netG)
    _test_postfilter_impl(model, model_config)



from functools import partial
from pathlib import Path

import hydra
import mlflow
import numpy as np
import torch
import torch.distributed as dist
from hydra.utils import to_absolute_path
from nnsvs.multistream import select_streams
from nnsvs.train_util import (
    collate_fn_default,
    collate_fn_random_segments,
    compute_distortions,
    eval_model,
    log_params_from_omegaconf_dict,
    save_checkpoint,
    save_configs,
    setup_gan,
)
from nnsvs.util import PyTorchStandardScaler, load_vocoder, make_non_pad_mask
from omegaconf import DictConfig
from torch import nn
from torch.cuda.amp import autocast
from torch.nn import functional as F


def train_step(
    model_config,
    optim_config,
    netG,
    optG,
    netD,
    optD,
    grad_scaler,
    train,
    in_feats,
    out_feats,
    lengths,
    out_scaler,
    mse_weight=1.0,
    adv_weight=1.0,
    adv_streams=None,
    fm_weight=0.0,
    mask_nth_mgc_for_adv_loss=0,
    gan_type="lsgan",
    vuv_mask=False,
):
    netG.train() if train else netG.eval()
    netD.train() if train else netD.eval()

    log_metrics = {}

    if vuv_mask:
        # NOTE: Assuming 3rd stream is the V/UV
        vuv_idx = np.sum(model_config.stream_sizes[:2])
        is_v = torch.logical_and(
            out_feats[:, :, vuv_idx : vuv_idx + 1] > 0,
            in_feats[:, :, vuv_idx : vuv_idx + 1] > 0,
        )
        vuv = is_v
    else:
        vuv = 1.0

    # Run forward
    with autocast(enabled=grad_scaler is not None):
        pred_out_feats = netG(in_feats, lengths)

    real_netD_in_feats = select_streams(
        out_feats, model_config.stream_sizes, adv_streams
    )
    fake_netD_in_feats = select_streams(
        pred_out_feats,
        model_config.stream_sizes,
        adv_streams,
    )

    # Ref: http://sython.org/papers/ASJ/saito2017asja.pdf
    # 0-th mgc with adversarial trainging affects speech quality
    # NOTE: assuming that the first stream contains mgc
    if mask_nth_mgc_for_adv_loss > 0:
        real_netD_in_feats = real_netD_in_feats[:, :, mask_nth_mgc_for_adv_loss:]
        fake_netD_in_feats = fake_netD_in_feats[:, :, mask_nth_mgc_for_adv_loss:]

    # Real
    with autocast(enabled=grad_scaler is not None):
        D_real = netD(real_netD_in_feats * vuv, in_feats, lengths)
        # NOTE: must be list of list to support multi-scale discriminators
        assert isinstance(D_real, list) and isinstance(D_real[-1], list)
        # Fake
        D_fake_det = netD(fake_netD_in_feats.detach() * vuv, in_feats, lengths)

    # Mask (B, T, 1)
    mask = make_non_pad_mask(lengths).unsqueeze(-1).to(in_feats.device)

    # Update discriminator
    eps = 1e-14
    loss_real = 0
    loss_fake = 0

    with autocast(enabled=grad_scaler is not None):
        for idx, (D_real_, D_fake_det_) in enumerate(zip(D_real, D_fake_det)):
            if gan_type == "lsgan":
                loss_real_ = (D_real_[-1] - 1) ** 2
                loss_fake_ = D_fake_det_[-1] ** 2
            elif gan_type == "vanilla-gan":
                loss_real_ = -torch.log(D_real_[-1] + eps)
                loss_fake_ = -torch.log(1 - D_fake_det_[-1] + eps)
            elif gan_type == "hinge":
                loss_real_ = F.relu(1 - D_real_[-1])
                loss_fake_ = F.relu(1 + D_fake_det_[-1])
            else:
                raise ValueError(f"Unknown gan type: {gan_type}")

            # mask for D
            if (
                hasattr(netD, "downsample_scale")
                and mask.shape[1] // netD.downsample_scale == D_real_[-1].shape[1]
            ):
                D_mask = mask[:, :: netD.downsample_scale, :]
            else:
                if D_real_[-1].shape[1] == out_feats.shape[1]:
                    D_mask = mask
                else:
                    D_mask = None

            if D_mask is not None:
                loss_real_ = loss_real_.masked_select(D_mask).mean()
                loss_fake_ = loss_fake_.masked_select(D_mask).mean()
            else:
                loss_real_ = loss_real_.mean()
                loss_fake_ = loss_fake_.mean()

            log_metrics[f"Loss_Real_Scale{idx}"] = loss_real_.item()
            log_metrics[f"Loss_Fake_Scale{idx}"] = loss_fake_.item()

            loss_real += loss_real_
            loss_fake += loss_fake_

        loss_d = loss_real + loss_fake

    if train:
        optD.zero_grad()
        if grad_scaler is not None:
            grad_scaler.scale(loss_d).backward()
            grad_scaler.unscale_(optD)
            grad_norm_d = torch.nn.utils.clip_grad_norm_(
                netD.parameters(), optim_config.netD.clip_norm
            )
            log_metrics["GradNorm_D"] = grad_norm_d
            grad_scaler.step(optD)
        else:
            loss_d.backward()
            grad_norm_d = torch.nn.utils.clip_grad_norm_(
                netD.parameters(), optim_config.netD.clip_norm
            )
            log_metrics["GradNorm_D"] = grad_norm_d
            optD.step()

    # MSE loss
    with autocast(enabled=grad_scaler is not None):
        loss_feats = nn.MSELoss(reduction="none")(
            pred_out_feats.masked_select(mask), out_feats.masked_select(mask)
        ).mean()

    # adversarial loss
    with autocast(enabled=grad_scaler is not None):
        D_fake = netD(fake_netD_in_feats * vuv, in_feats, lengths)

        loss_adv = 0
        for idx, D_fake_ in enumerate(D_fake):
            if gan_type == "lsgan":
                loss_adv_ = (1 - D_fake_[-1]) ** 2
            elif gan_type == "vanilla-gan":
                loss_adv_ = -torch.log(D_fake_[-1] + eps)
            elif gan_type == "hinge":
                loss_adv_ = -D_fake_[-1]
            else:
                raise ValueError(f"Unknown gan type: {gan_type}")

            if (
                hasattr(netD, "downsample_scale")
                and mask.shape[1] // netD.downsample_scale == D_fake_[-1].shape[1]
            ):
                D_mask = mask[:, :: netD.downsample_scale, :]
            else:
                if D_real_[-1].shape[1] == out_feats.shape[1]:
                    D_mask = mask
                else:
                    D_mask = None

            if D_mask is not None:
                loss_adv_ = loss_adv_.masked_select(D_mask).mean()
            else:
                loss_adv_ = loss_adv_.mean()

            log_metrics[f"Loss_Adv_Scale{idx}"] = loss_adv_.item()

            loss_adv += loss_adv_

        # Feature matching loss
        loss_fm = torch.tensor(0.0).to(in_feats.device)
        if fm_weight > 0:
            for D_fake_, D_real_ in zip(D_fake, D_real):
                for fake_fmap, real_fmap in zip(D_fake_[:-1], D_real_[:-1]):
                    loss_fm += F.l1_loss(fake_fmap, real_fmap.detach())

        loss = mse_weight * loss_feats + adv_weight * loss_adv + fm_weight * loss_fm

    if train:
        optG.zero_grad()
        if grad_scaler is not None:
            grad_scaler.scale(loss).backward()
            grad_scaler.unscale_(optG)
            grad_norm_g = torch.nn.utils.clip_grad_norm_(
                netG.parameters(), optim_config.netG.clip_norm
            )
            log_metrics["GradNorm_G"] = grad_norm_g
            grad_scaler.step(optG)
        else:
            loss.backward()
            grad_norm_g = torch.nn.utils.clip_grad_norm_(
                netG.parameters(), optim_config.netG.clip_norm
            )
            log_metrics["GradNorm_G"] = grad_norm_g
            optG.step()

    # NOTE: this shouldn't be called multiple times in a training step
    if train and grad_scaler is not None:
        grad_scaler.update()

    # Metrics
    distortions = compute_distortions(
        pred_out_feats, out_feats, lengths, out_scaler, model_config
    )
    log_metrics.update(distortions)
    log_metrics.update(
        {
            "Loss": loss.item(),
            "Loss_Feats": loss_feats.item(),
            "Loss_Adv_Total": loss_adv.item(),
            "Loss_Feature_Matching": loss_fm.item(),
            "Loss_Real_Total": loss_real.item(),
            "Loss_Fake_Total": loss_fake.item(),
            "Loss_D": loss_d.item(),
        }
    )

    return loss, log_metrics


def train_loop(
    config,
    logger,
    device,
    netG,
    optG,
    schedulerG,
    netD,
    optD,
    schedulerD,
    grad_scaler,
    data_loaders,
    samplers,
    writer,
    out_scaler,
    use_mlflow,
    vocoder,
    vocoder_in_scaler,
):
    out_dir = Path(to_absolute_path(config.train.out_dir))
    best_dev_loss = torch.finfo(torch.float32).max
    last_dev_loss = torch.finfo(torch.float32).max

    adv_streams = config.train.adv_streams
    if len(adv_streams) != len(config.model.stream_sizes):
        raise ValueError("adv_streams must be specified for all streams")

    if dist.is_initialized() and dist.get_rank() != 0:

        def tqdm(x, **kwargs):
            return x

    else:
        from tqdm import tqdm

    train_iter = 1
    for epoch in tqdm(range(1, config.train.nepochs + 1)):
        for phase in data_loaders.keys():
            train = phase.startswith("train")
            # https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
            if dist.is_initialized() and train and samplers[phase] is not None:
                samplers[phase].set_epoch(epoch)
            running_loss = 0
            running_metrics = {}
            evaluated = False
            for in_feats, out_feats, lengths in tqdm(
                data_loaders[phase], desc=f"{phase} iter", leave=False
            ):
                # NOTE: This is needed for pytorch's PackedSequence
                lengths, indices = torch.sort(lengths, dim=0, descending=True)
                in_feats, out_feats = (
                    in_feats[indices].to(device),
                    out_feats[indices].to(device),
                )
                if (not train) and (not evaluated):
                    eval_model(
                        phase,
                        epoch,
                        netG,
                        in_feats,
                        out_feats,
                        lengths,
                        config.model,
                        out_scaler,
                        writer,
                        sr=config.data.sample_rate,
                        use_world_codec=config.data.use_world_codec,
                        vocoder=vocoder,
                        vocoder_in_scaler=vocoder_in_scaler,
                        max_num_eval_utts=config.train.max_num_eval_utts,
                    )
                    evaluated = True

                loss, log_metrics = train_step(
                    model_config=config.model,
                    optim_config=config.train.optim,
                    netG=netG,
                    optG=optG,
                    netD=netD,
                    optD=optD,
                    grad_scaler=grad_scaler,
                    train=train,
                    in_feats=in_feats,
                    out_feats=out_feats,
                    lengths=lengths,
                    out_scaler=out_scaler,
                    mse_weight=config.train.mse_weight,
                    adv_weight=config.train.adv_weight,
                    adv_streams=adv_streams,
                    fm_weight=config.train.fm_weight,
                    mask_nth_mgc_for_adv_loss=config.train.mask_nth_mgc_for_adv_loss,
                    gan_type=config.train.gan_type,
                    vuv_mask=config.train.vuv_mask,
                )

                if train:
                    if writer is not None:
                        for key, val in log_metrics.items():
                            writer.add_scalar(f"{key}_Step/{phase}", val, train_iter)
                    train_iter += 1

                running_loss += loss.item()
                for k, v in log_metrics.items():
                    try:
                        running_metrics[k] += float(v)
                    except KeyError:
                        running_metrics[k] = float(v)

            ave_loss = running_loss / len(data_loaders[phase])
            logger.info("[%s] [Epoch %s]: loss %s", phase, epoch, ave_loss)
            if writer is not None:
                writer.add_scalar(f"Loss_Epoch/{phase}", ave_loss, epoch)
            if use_mlflow:
                mlflow.log_metric(f"{phase}_loss", ave_loss, step=epoch)

            for k, v in running_metrics.items():
                ave_v = v / len(data_loaders[phase])
                if writer is not None:
                    writer.add_scalar(f"{k}_Epoch/{phase}", ave_v, epoch)
                if use_mlflow:
                    mlflow.log_metric(f"{phase}_{k}", ave_v, step=epoch)

            if not train:
                last_dev_loss = ave_loss
            if not train and ave_loss < best_dev_loss:
                best_dev_loss = ave_loss
                for model, opt, scheduler, postfix in [
                    (netG, optG, schedulerG, ""),
                    (netD, optD, schedulerD, "_D"),
                ]:
                    save_checkpoint(
                        logger,
                        out_dir,
                        model,
                        opt,
                        scheduler,
                        epoch,
                        is_best=True,
                        postfix=postfix,
                    )

        schedulerG.step()
        schedulerD.step()

        if epoch % config.train.checkpoint_epoch_interval == 0:
            for model, opt, scheduler, postfix in [
                (netG, optG, schedulerG, ""),
                (netD, optD, schedulerD, "_D"),
            ]:
                save_checkpoint(
                    logger,
                    out_dir,
                    model,
                    opt,
                    scheduler,
                    epoch,
                    is_best=False,
                    postfix=postfix,
                )

    for model, opt, scheduler, postfix in [
        (netG, optG, schedulerG, ""),
        (netD, optD, schedulerD, "_D"),
    ]:
        save_checkpoint(
            logger,
            out_dir,
            model,
            opt,
            scheduler,
            config.train.nepochs,
            postfix=postfix,
        )
    logger.info("The best loss was %s", best_dev_loss)
    if use_mlflow:
        mlflow.log_metric("best_dev_loss", best_dev_loss, step=epoch)
        mlflow.log_artifacts(out_dir)

    return last_dev_loss


@hydra.main(config_path="conf/train_postfilter", config_name="config")
def my_app(config: DictConfig) -> None:
    # NOTE: set discriminator's in_dim automatically
    if config.model.netD.in_dim is None:
        stream_sizes = np.asarray(config.model.stream_sizes)
        D_in_dim = int((stream_sizes * np.asarray(config.train.adv_streams)).sum())
        if config.train.mask_nth_mgc_for_adv_loss > 0:
            D_in_dim -= config.train.mask_nth_mgc_for_adv_loss
        config.model.netD.in_dim = D_in_dim
    if "stream_sizes" in config.model.netG:
        config.model.netG.stream_sizes = config.model.stream_sizes

    if "max_time_frames" in config.data and config.data.max_time_frames > 0:
        collate_fn = partial(
            collate_fn_random_segments, max_time_frames=config.data.max_time_frames
        )
    else:
        collate_fn = collate_fn_default

    if config.train.use_ddp:
        dist.init_process_group("nccl")
        rank = dist.get_rank()
        device_id = rank % torch.cuda.device_count()
        torch.cuda.set_device(device_id)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    (
        (netG, optG, schedulerG),
        (netD, optD, schedulerD),
        grad_scaler,
        data_loaders,
        samplers,
        writer,
        logger,
        _,
        out_scaler,
    ) = setup_gan(config, device, collate_fn)

    path = config.train.pretrained_vocoder_checkpoint
    if path is not None and len(path) > 0:
        logger.info(f"Loading pretrained vocoder checkpoint from {path}")
        vocoder, vocoder_in_scaler = load_vocoder(path, device)
    else:
        vocoder, vocoder_in_scaler = None, None

    out_scaler = PyTorchStandardScaler(
        torch.from_numpy(out_scaler.mean_), torch.from_numpy(out_scaler.scale_)
    ).to(device)
    use_mlflow = config.mlflow.enabled

    if use_mlflow:
        with mlflow.start_run() as run:
            # NOTE: modify out_dir when running with mlflow
            config.train.out_dir = f"{config.train.out_dir}/{run.info.run_id}"
            save_configs(config)
            log_params_from_omegaconf_dict(config)
            last_dev_loss = train_loop(
                config,
                logger,
                device,
                netG,
                optG,
                schedulerG,
                netD,
                optD,
                schedulerD,
                grad_scaler,
                data_loaders,
                samplers,
                writer,
                out_scaler,
                use_mlflow,
                vocoder,
                vocoder_in_scaler,
            )
    else:
        save_configs(config)
        last_dev_loss = train_loop(
            config,
            logger,
            device,
            netG,
            optG,
            schedulerG,
            netD,
            optD,
            schedulerD,
            grad_scaler,
            data_loaders,
            samplers,
            writer,
            out_scaler,
            use_mlflow,
            vocoder,
            vocoder_in_scaler,
        )

    return last_dev_loss


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



import importlib
import random
from os.path import join
from pathlib import Path
from typing import Any

import numpy as np
import pkg_resources
import pyworld
import torch
from hydra.utils import instantiate
from nnsvs.multistream import get_static_features, get_static_stream_sizes
from nnsvs.usfgan import USFGANWrapper
from omegaconf import OmegaConf
from torch import nn

try:
    from parallel_wavegan.utils import load_model

    _pwg_available = True
except ImportError:
    _pwg_available = False


# mask-related functions were adapted from https://github.com/espnet/espnet

EXAMPLE_DIR = "_example_data"


# Adapted from https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix
def init_weights(net, init_type="normal", init_gain=0.02):
    """Initialize network weights.

    Args:
        net (torch.nn.Module): network to initialize
        init_type (str): the name of an initialization method:
            normal | xavier | kaiming | orthogonal | none.
        init_gain (float): scaling factor for normal, xavier and orthogonal.
    """
    if init_type == "none":
        return

    def init_func(m):
        classname = m.__class__.__name__
        if hasattr(m, "weight") and (
            classname.find("Conv") != -1 or classname.find("Linear") != -1
        ):
            if init_type == "normal":
                nn.init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == "xavier_normal":
                nn.init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == "kaiming_normal":
                nn.init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
            elif init_type == "orthogonal":
                nn.init.orthogonal_(m.weight.data, gain=init_gain)
            else:
                raise NotImplementedError(
                    "initialization method [%s] is not implemented" % init_type
                )
            if hasattr(m, "bias") and m.bias is not None:
                nn.init.constant_(m.bias.data, 0.0)
        elif classname.find("BatchNorm2d") != -1:
            # BatchNorm Layer's weight is not a matrix; only normal distribution applies.
            nn.init.normal_(m.weight.data, 1.0, init_gain)
            nn.init.constant_(m.bias.data, 0.0)

    net.apply(init_func)


def get_world_stream_info(
    sr: int,
    mgc_order: int,
    num_windows: int = 3,
    vibrato_mode: str = "none",
    use_mcep_aperiodicity: bool = False,
    mcep_aperiodicity_order: int = 24,
):
    """Get stream sizes for WORLD-based acoustic features

    Args:
        sr (int): sampling rate
        mgc_order (int): order of mel-generalized cepstrum
        num_windows (int): number of windows
        vibrato_mode (str): vibrato analysis mode

    Returns:
        list: stream sizes
    """
    # [mgc, lf0, vuv, bap]
    stream_sizes = [
        (mgc_order + 1) * num_windows,
        num_windows,
        1,
        pyworld.get_num_aperiodicities(sr) * num_windows
        if not use_mcep_aperiodicity
        else mcep_aperiodicity_order + 1,
    ]
    if vibrato_mode == "diff":
        # vib
        stream_sizes.append(num_windows)
    elif vibrato_mode == "sine":
        # vib + vib_flags
        stream_sizes.append(3 * num_windows)
        stream_sizes.append(1)
    elif vibrato_mode == "none":
        pass
    else:
        raise RuntimeError("Unknown vibrato mode: {}".format(vibrato_mode))

    return stream_sizes


def load_utt_list(utt_list):
    """Load a list of utterances.

    Args:
        utt_list (str): path to a file containing a list of utterances

    Returns:
        List[str]: list of utterances
    """
    with open(utt_list) as f:
        utt_ids = f.readlines()
    utt_ids = map(lambda utt_id: utt_id.strip(), utt_ids)
    utt_ids = filter(lambda utt_id: len(utt_id) > 0, utt_ids)
    return list(utt_ids)


def example_xml_file(key="haruga_kita"):
    """Get the path to an included xml file.

    Args:
        key (str): key of the file

    Returns:
        str: path to an example xml file

    Raises:
        FileNotFoundError: if the file is not found
    """
    return pkg_resources.resource_filename(__name__, join(EXAMPLE_DIR, f"{key}.xml"))


def init_seed(seed):
    """Initialize random seed.

    Args:
        seed (int): random seed
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


def dynamic_import(name: str) -> Any:
    """Dynamic import

    Args:
        name (str): module_name + ":" + class_name

    Returns:
        Any: class object
    """
    mod_name, class_name = name.split(":")
    mod = importlib.import_module(mod_name)
    return getattr(mod, class_name)


def pad_2d(x, max_len, constant_values=0):
    """Pad a 2d-tensor.

    Args:
        x (torch.Tensor): tensor to pad
        max_len (int): maximum length of the tensor
        constant_values (int, optional): value to pad with. Default: 0

    Returns:
        torch.Tensor: padded tensor
    """
    x = np.pad(
        x,
        [(0, max_len - len(x)), (0, 0)],
        mode="constant",
        constant_values=constant_values,
    )
    return x


def make_pad_mask(lengths, xs=None, length_dim=-1, maxlen=None):
    """Make mask tensor containing indices of padded part.

    Args:
        lengths (LongTensor or List): Batch of lengths (B,).
        xs (Tensor, optional): The reference tensor. If set, masks will be
            the same shape as this tensor.
        length_dim (int, optional): Dimension indicator of the above tensor.

    Returns:
        Tensor: Mask tensor containing indices of padded part.
                dtype=torch.uint8 in PyTorch 1.2-
                dtype=torch.bool in PyTorch 1.2+ (including 1.2)
    """
    if length_dim == 0:
        raise ValueError("length_dim cannot be 0: {}".format(length_dim))

    if not isinstance(lengths, list):
        lengths = lengths.tolist()
    bs = int(len(lengths))
    if maxlen is None:
        if xs is None:
            maxlen = int(max(lengths))
        else:
            maxlen = xs.size(length_dim)

    seq_range = torch.arange(0, maxlen, dtype=torch.int64)
    seq_range_expand = seq_range.unsqueeze(0).expand(bs, maxlen)
    seq_length_expand = seq_range_expand.new(lengths).unsqueeze(-1)
    mask = seq_range_expand >= seq_length_expand

    if xs is not None:
        assert xs.size(0) == bs, (xs.size(0), bs)

        if length_dim < 0:
            length_dim = xs.dim() + length_dim
        # ind = (:, None, ..., None, :, , None, ..., None)
        ind = tuple(
            slice(None) if i in (0, length_dim) else None for i in range(xs.dim())
        )
        mask = mask[ind].expand_as(xs).to(xs.device)
    return mask


def make_non_pad_mask(lengths, xs=None, length_dim=-1, maxlen=None):
    """Make mask tensor containing indices of non-padded part.

    Args:
        lengths (LongTensor or List): Batch of lengths (B,).
        xs (Tensor, optional): The reference tensor.
            If set, masks will be the same shape as this tensor.
        length_dim (int, optional): Dimension indicator of the above tensor.

    Returns:
        ByteTensor: mask tensor containing indices of padded part.
                    dtype=torch.uint8 in PyTorch 1.2-
                    dtype=torch.bool in PyTorch 1.2+ (including 1.2)
    """
    return ~make_pad_mask(lengths, xs, length_dim, maxlen)


class PyTorchStandardScaler(nn.Module):
    """PyTorch module for standardization.

    Args:
        mean (torch.Tensor): mean
        scale (torch.Tensor): scale
    """

    def __init__(self, mean, scale):
        super().__init__()
        self.mean_ = nn.Parameter(mean, requires_grad=False)
        self.scale_ = nn.Parameter(scale, requires_grad=False)

    def transform(self, x):
        return (x - self.mean_) / self.scale_

    def inverse_transform(self, x):
        return x * self.scale_ + self.mean_


class StandardScaler:
    """sklearn.preprocess.StandardScaler like class with only
    transform functionality

    Args:
        mean (np.ndarray): mean
        var (np.ndarray): variance
        scale (np.ndarray): scale
    """

    def __init__(self, mean, var, scale):
        self.mean_ = mean
        self.var_ = var
        # NOTE: scale may not exactly same as np.sqrt(var)
        self.scale_ = scale

    def transform(self, x):
        return (x - self.mean_) / self.scale_

    def inverse_transform(self, x):
        return x * self.scale_ + self.mean_


class MinMaxScaler:
    """sklearn.preprocess.MinMaxScaler like class with only
    transform functionality

    Args:
        min (np.ndarray): minimum
        scale (np.ndarray): scale
        data_min (np.ndarray): minimum of input data
        data_max (np.ndarray): maximum of input data
        feature_range (tuple): (min, max)
    """

    def __init__(self, min, scale, data_min=None, data_max=None, feature_range=(0, 1)):
        self.min_ = min
        self.scale_ = scale
        self.data_min_ = data_min
        self.data_max_ = data_max
        self.feature_range = feature_range

    def transform(self, x):
        return self.scale_ * x + self.min_

    def inverse_transform(self, x):
        return (x - self.min_) / self.scale_


def extract_static_scaler(out_scaler, model_config):
    """Extract scaler for static features

    Args:
        out_scaler (StandardScaler or MinMaxScaler): target scaler
        model_config (dict): model config that contain stream information

    Returns:
        StandardScaler or MinMaxScaler: scaler for static features
    """
    mean_ = get_static_features(
        out_scaler.mean_.reshape(1, 1, out_scaler.mean_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    mean_ = np.concatenate(mean_, -1).reshape(1, -1)
    var_ = get_static_features(
        out_scaler.var_.reshape(1, 1, out_scaler.var_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    var_ = np.concatenate(var_, -1).reshape(1, -1)
    scale_ = get_static_features(
        out_scaler.scale_.reshape(1, 1, out_scaler.scale_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    scale_ = np.concatenate(scale_, -1).reshape(1, -1)
    static_scaler = StandardScaler(mean_, var_, scale_)
    return static_scaler


def load_vocoder(path, device, acoustic_config):
    """Load vocoder model from a given checkpoint path

    Note that the path needs to be a checkpoint of PWG or USFGAN.

    Args:
        path (str or Path): Path to the vocoder model
        device (str): Device to load the model
        acoustic_config (dict): Acoustic model config

    Returns:
        tuple: (vocoder, vocoder_in_scaler, vocoder_config)
    """
    if not _pwg_available:
        raise RuntimeError(
            "parallel_wavegan is required to load pre-trained checkpoint."
        )
    path = Path(path) if isinstance(path, str) else path
    model_dir = path.parent
    if (model_dir / "vocoder_model.yaml").exists():
        # packed model
        vocoder_config = OmegaConf.load(model_dir / "vocoder_model.yaml")
    elif (model_dir / "config.yml").exists():
        # PWG checkpoint
        vocoder_config = OmegaConf.load(model_dir / "config.yml")
    else:
        # usfgan
        vocoder_config = OmegaConf.load(model_dir / "config.yaml")

    if "generator" in vocoder_config and "discriminator" in vocoder_config:
        # usfgan
        checkpoint = torch.load(
            path,
            map_location=lambda storage, loc: storage,
        )

        vocoder = instantiate(vocoder_config.generator).to(device)
        vocoder.load_state_dict(checkpoint["model"]["generator"])
        vocoder.remove_weight_norm()
        vocoder = USFGANWrapper(vocoder_config, vocoder)

        stream_sizes = get_static_stream_sizes(
            acoustic_config.stream_sizes,
            acoustic_config.has_dynamic_features,
            acoustic_config.num_windows,
        )

        # Extract scaler params for [mgc, bap]
        if vocoder_config.data.aux_feats == ["mcep", "codeap"]:
            # streams: (mgc, lf0, vuv, bap)
            mean_ = np.load(model_dir / "in_vocoder_scaler_mean.npy")
            var_ = np.load(model_dir / "in_vocoder_scaler_var.npy")
            scale_ = np.load(model_dir / "in_vocoder_scaler_scale.npy")
            mgc_end_dim = stream_sizes[0]
            bap_start_dim = sum(stream_sizes[:3])
            bap_end_dim = sum(stream_sizes[:4])
            vocoder_in_scaler = StandardScaler(
                np.concatenate([mean_[:mgc_end_dim], mean_[bap_start_dim:bap_end_dim]]),
                np.concatenate([var_[:mgc_end_dim], var_[bap_start_dim:bap_end_dim]]),
                np.concatenate(
                    [scale_[:mgc_end_dim], scale_[bap_start_dim:bap_end_dim]]
                ),
            )
        else:
            # streams: (mel, lf0, vuv)
            mel_dim = stream_sizes[0]
            vocoder_in_scaler = StandardScaler(
                np.load(model_dir / "in_vocoder_scaler_mean.npy")[:mel_dim],
                np.load(model_dir / "in_vocoder_scaler_var.npy")[:mel_dim],
                np.load(model_dir / "in_vocoder_scaler_scale.npy")[:mel_dim],
            )
    else:
        # PWG
        vocoder = load_model(path, config=vocoder_config).to(device)
        vocoder.remove_weight_norm()
        vocoder_in_scaler = StandardScaler(
            np.load(model_dir / "in_vocoder_scaler_mean.npy"),
            np.load(model_dir / "in_vocoder_scaler_var.npy"),
            np.load(model_dir / "in_vocoder_scaler_scale.npy"),
        )
    vocoder.eval()

    return vocoder, vocoder_in_scaler, vocoder_config



import numpy as np
import torch
from nnsvs.base import BaseModel
from nnsvs.multistream import split_streams
from nnsvs.util import init_weights
from torch import nn


def variance_scaling(gv, feats, offset=2, note_frame_indices=None):
    """Variance scaling method to enhance synthetic speech quality

    Method proposed in :cite:t:`silen2012ways`.

    Args:
        gv (tensor): global variance computed over training data
        feats (tensor): input features
        offset (int): offset
        note_frame_indices (tensor): indices of note frames

    Returns:
        tensor: scaled features
    """
    if note_frame_indices is not None:
        if len(note_frame_indices) == 0:
            return feats
        utt_gv = feats[note_frame_indices].var(0)
        utt_mu = feats[note_frame_indices].mean(0)
    else:
        utt_gv = feats.var(0)
        utt_mu = feats.mean(0)

    out = feats.copy()
    if note_frame_indices is not None:
        out[note_frame_indices, offset:] = (
            np.sqrt(gv[offset:] / utt_gv[offset:])
            * (feats[note_frame_indices, offset:] - utt_mu[offset:])
            + utt_mu[offset:]
        )
    else:
        out[:, offset:] = (
            np.sqrt(gv[offset:] / utt_gv[offset:])
            * (feats[:, offset:] - utt_mu[offset:])
            + utt_mu[offset:]
        )

    return out


class MovingAverage1d(nn.Conv1d):
    """Moving average filter on 1-D signals

    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output channels
        kernel_size (int): kernel size
        padding_mode (str): padding mode
    """

    def __init__(self, in_channels, out_channels, kernel_size, padding_mode="reflect"):
        # NOTE: process each channel independently by setting groups=in_channels
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            groups=in_channels,
            bias=False,
            padding="same",
            padding_mode=padding_mode,
        )
        nn.init.constant_(self.weight, 1 / kernel_size)
        for p in self.parameters():
            p.requires_grad = False


class Conv2dPostFilter(BaseModel):
    """A post-filter based on Conv2d

    A model proposed in :cite:t:`kaneko2017generative`.

    Args:
        channels (int): number of channels
        kernel_size (tuple): kernel sizes for Conv2d
        init_type (str): type of initialization
        noise_scale (float): scale of noise
        noise_type (str): type of noise. "frame_wise" or "bin_wise"
        padding_mode (str): padding mode
        smoothing_width (int): Width of smoothing window.
            The larger the smoother. Only used at inference time.
    """

    def __init__(
        self,
        in_dim=None,
        channels=128,
        kernel_size=(5, 5),
        init_type="kaiming_normal",
        noise_scale=1.0,
        noise_type="bin_wise",
        padding_mode="zeros",
        smoothing_width=-1,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.noise_type = noise_type
        self.noise_scale = noise_scale
        C = channels
        self.smoothing_width = smoothing_width
        assert len(kernel_size) == 2
        ks = np.asarray(list(kernel_size))
        padding = (ks - 1) // 2
        self.conv1 = nn.Sequential(
            nn.Conv2d(
                2,
                C,
                kernel_size=ks,
                padding=padding,
                padding_mode=padding_mode,
            ),
            nn.ReLU(),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(
                C + 1, C * 2, kernel_size=ks, padding=padding, padding_mode=padding_mode
            ),
            nn.ReLU(),
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(
                C * 2 + 1, C, kernel_size=ks, padding=padding, padding_mode=padding_mode
            ),
            nn.ReLU(),
        )
        self.conv4 = nn.Conv2d(
            C + 1, 1, kernel_size=ks, padding=padding, padding_mode=padding_mode
        )

        if self.noise_type == "frame_wise":
            # noise: (B, T, 1)
            self.fc = nn.Linear(1, in_dim)
        elif self.noise_type == "bin_wise":
            # noise: (B, T, C)
            self.fc = None
        else:
            raise ValueError("Unknown noise type: {}".format(self.noise_type))

        init_weights(self, init_type)

    def forward(self, x, lengths=None, y=None, is_inference=False):
        """Forward step

        Args:
            x (torch.Tensor): input tensor of shape (B, T, C)
            lengths (torch.Tensor): lengths of shape (B,)

        Returns:
            torch.Tensor: output tensor of shape (B, T, C)
        """
        # (B, T, C) -> (B, 1, T, C):
        x = x.unsqueeze(1)
        if self.noise_type == "bin_wise":
            # (B, C, T)
            z = torch.randn_like(x).squeeze(1).transpose(1, 2) * self.noise_scale
            # Apply moving average filter at inference time only
            if is_inference and self.smoothing_width > 0:
                ave_filt = MovingAverage1d(
                    self.in_dim, self.in_dim, self.smoothing_width
                ).to(x.device)
                z = ave_filt(z)
            # (B, 1, T, C)
            z = z.transpose(1, 2).unsqueeze(1)
        elif self.noise_type == "frame_wise":
            # (B, 1, T)
            z = torch.randn(x.shape[0], 1, x.shape[2]).to(x.device) * self.noise_scale
            # Apply moving average filter at inference time only
            if is_inference and self.smoothing_width > 0:
                ave_filt = MovingAverage1d(1, 1, self.smoothing_width).to(x.device)
                z = ave_filt(z)
            # (B, 1, T, 1)
            z = z.unsqueeze(-1)
            # (B, 1, T, C)
            z = self.fc(z)

        x_syn = x

        y = self.conv1(torch.cat([x_syn, z], dim=1))
        y = self.conv2(torch.cat([x_syn, y], dim=1))
        y = self.conv3(torch.cat([x_syn, y], dim=1))
        residual = self.conv4(torch.cat([x_syn, y], dim=1))
        out = x_syn + residual

        # (B, 1, T, C) -> (B, T, C)
        out = out.squeeze(1)

        return out

    def inference(self, x, lengths=None):
        return self(x, lengths, is_inference=True)


class MultistreamPostFilter(BaseModel):
    """A multi-stream post-filter that applies post-filtering for each feature stream

    Currently, post-filtering for MGC, BAP and log-F0 are supported.
    Note that it doesn't make much sense to apply post-filtering for other features.

    Args:
        mgc_postfilter (nn.Module): post-filter for MGC
        bap_postfilter (nn.Module): post-filter for BAP
        lf0_postfilter (nn.Module): post-filter for log-F0
        stream_sizes (list): sizes of each feature stream
        mgc_offset (int): offset for MGC. Defaults to 2.
        bap_offset (int): offset for BAP. Defaults to 0.
    """

    def __init__(
        self,
        mgc_postfilter: nn.Module,
        bap_postfilter: nn.Module,
        lf0_postfilter: nn.Module,
        stream_sizes: list,
        mgc_offset: int = 2,
        bap_offset: int = 0,
    ):
        super().__init__()
        self.mgc_postfilter = mgc_postfilter
        self.bap_postfilter = bap_postfilter
        self.lf0_postfilter = lf0_postfilter
        self.stream_sizes = stream_sizes
        self.mgc_offset = mgc_offset
        self.bap_offset = bap_offset

    def forward(self, x, lengths=None, y=None, is_inference=False):
        """Forward step

        Each feature stream is processed independently.

        Args:
            x (torch.Tensor): input tensor of shape (B, T, C)
            lengths (torch.Tensor): lengths of shape (B,)

        Returns:
            torch.Tensor: output tensor of shape (B, T, C)
        """
        streams = split_streams(x, self.stream_sizes)
        if len(streams) == 4:
            mgc, lf0, vuv, bap = streams
        elif len(streams) == 5:
            mgc, lf0, vuv, bap, vuv = streams
        elif len(streams) == 6:
            mgc, lf0, vuv, bap, vib, vib_flags = streams
        else:
            raise ValueError("Invalid number of streams")

        if self.mgc_postfilter is not None:
            if self.mgc_offset > 0:
                # keep unchanged for the 0-to-${mgc_offset}-th dim of mgc
                mgc0 = mgc[:, :, : self.mgc_offset]
                if is_inference:
                    mgc_pf = self.mgc_postfilter.inference(
                        mgc[:, :, self.mgc_offset :], lengths
                    )
                else:
                    mgc_pf = self.mgc_postfilter(mgc[:, :, self.mgc_offset :], lengths)
                mgc_pf = torch.cat([mgc0, mgc_pf], dim=-1)
            else:
                if is_inference:
                    mgc_pf = self.mgc_postfilter.inference(mgc, lengths)
                else:
                    mgc_pf = self.mgc_postfilter(mgc, lengths)
            mgc = mgc_pf

        if self.bap_postfilter is not None:
            if self.bap_offset > 0:
                # keep unchanged for the 0-to-${bap_offset}-th dim of bap
                bap0 = bap[:, :, : self.bap_offset]
                if is_inference:
                    bap_pf = self.bap_postfilter.inference(
                        bap[:, :, self.bap_offset :], lengths
                    )
                else:
                    bap_pf = self.bap_postfilter(bap[:, :, self.bap_offset :], lengths)
                bap_pf = torch.cat([bap0, bap_pf], dim=-1)
            else:
                if is_inference:
                    bap_pf = self.bap_postfilter.inference(bap, lengths)
                else:
                    bap_pf = self.bap_postfilter(bap, lengths)
            bap = bap_pf

        if self.lf0_postfilter is not None:
            if is_inference:
                lf0 = self.lf0_postfilter.inference(lf0, lengths)
            else:
                lf0 = self.lf0_postfilter(lf0, lengths)

        if len(streams) == 4:
            out = torch.cat([mgc, lf0, vuv, bap], dim=-1)
        elif len(streams) == 5:
            out = torch.cat([mgc, lf0, vuv, bap, vib], dim=-1)
        elif len(streams) == 6:
            out = torch.cat([mgc, lf0, vuv, bap, vib, vib_flags], dim=-1)

        return out

    def inference(self, x, lengths):
        return self(x, lengths, is_inference=True)


class MelF0MultistreamPostFilter(BaseModel):
    def __init__(
        self,
        mel_postfilter: nn.Module,
        lf0_postfilter: nn.Module,
        stream_sizes: list,
        mel_offset: int = 0,
    ):
        super().__init__()
        self.mel_postfilter = mel_postfilter
        self.lf0_postfilter = lf0_postfilter
        self.stream_sizes = stream_sizes
        self.mel_offset = mel_offset

    def forward(self, x, lengths=None, y=None, is_inference=False):
        """Forward step

        Each feature stream is processed independently.

        Args:
            x (torch.Tensor): input tensor of shape (B, T, C)
            lengths (torch.Tensor): lengths of shape (B,)

        Returns:
            torch.Tensor: output tensor of shape (B, T, C)
        """
        streams = split_streams(x, self.stream_sizes)
        assert len(streams) == 3
        mel, lf0, vuv = streams

        if self.mel_postfilter is not None:
            if self.mel_offset > 0:
                # keep unchanged for the 0-to-${mgc_offset}-th dim of mgc
                mel0 = mel[:, :, : self.mel_offset]
                if is_inference:
                    mel_pf = self.mel_postfilter.inference(
                        mel[:, :, self.mel_offset :], lengths
                    )
                else:
                    mel_pf = self.mel_postfilter(mel[:, :, self.mel_offset :], lengths)
                mel_pf = torch.cat([mel0, mel_pf], dim=-1)
            else:
                if is_inference:
                    mel_pf = self.mel_postfilter.inference(mel, lengths)
                else:
                    mel_pf = self.mel_postfilter(mel, lengths)
            mel = mel_pf

        if self.lf0_postfilter is not None:
            if is_inference:
                lf0 = self.lf0_postfilter.inference(lf0, lengths)
            else:
                lf0 = self.lf0_postfilter(lf0, lengths)

        out = torch.cat([mel, lf0, vuv], dim=-1)

        return out

    def inference(self, x, lengths):
        return self(x, lengths, is_inference=True)


class _PadConv2dPostFilter(BaseModel):
    def __init__(
        self,
        in_dim=None,
        channels=128,
        kernel_size=5,
        init_type="kaiming_normal",
        padding_side="left",
    ):
        super().__init__()
        assert not isinstance(kernel_size, list)
        C = channels
        ks = kernel_size
        padding = (ks - 1) // 2
        self.padding = padding

        # Treat padding for the feature-axis carefully
        # use normal padding for the time-axis (i.e., (padding, padding))
        self.padding_side = padding_side
        if padding_side == "left":
            self.pad = nn.ReflectionPad2d((padding, 0, padding, padding))
        elif padding_side == "none":
            self.pad = nn.ReflectionPad2d((0, 0, padding, padding))
        elif padding_side == "right":
            self.pad = nn.ReflectionPad2d((0, padding, padding, padding))
        else:
            raise ValueError("Invalid padding side")

        self.conv1 = nn.Sequential(
            nn.Conv2d(2, C, kernel_size=(ks, ks)),
            nn.ReLU(),
        )
        # NOTE: for the subsequent layers, use fixed kernel_size 3 for feature-axis
        self.conv2 = nn.Sequential(
            nn.Conv2d(
                C + 1,
                C * 2,
                kernel_size=(ks, 3),
                padding=(padding, 1),
                padding_mode="reflect",
            ),
            nn.ReLU(),
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(
                C * 2 + 1,
                C,
                kernel_size=(ks, 3),
                padding=(padding, 1),
                padding_mode="reflect",
            ),
            nn.ReLU(),
        )
        self.conv4 = nn.Conv2d(
            C + 1, 1, kernel_size=(ks, 1), padding=(padding, 0), padding_mode="reflect"
        )

        self.fc = nn.Linear(1, in_dim)

        init_weights(self, init_type)

    def forward(self, x, z, lengths=None):
        # (B, T, C) -> (B, 1, T, C):
        x = x.unsqueeze(1)
        z = z.unsqueeze(1)

        z = self.fc(z)

        x_syn = x
        y = self.conv1(torch.cat([self.pad(x_syn), self.pad(z)], dim=1))

        if self.padding_side == "left":
            x_syn = x[:, :, :, : -self.padding]
        elif self.padding_side == "none":
            x_syn = x[:, :, :, self.padding : -self.padding]
        elif self.padding_side == "right":
            x_syn = x[:, :, :, self.padding :]

        y = self.conv2(torch.cat([x_syn, y], dim=1))
        y = self.conv3(torch.cat([x_syn, y], dim=1))
        residual = self.conv4(torch.cat([x_syn, y], dim=1))
        out = x_syn + residual

        # (B, 1, T, C) -> (B, T, C)
        out = out.squeeze(1)

        return out


class MultistreamConv2dPostFilter(nn.Module):
    """Conv2d-based multi-stream post-filter designed for MGC

    Divide the MGC transformation into low/mid/high dim transfomations
    with small overlaps. Overlap is determined by the kernel size.
    """

    def __init__(
        self,
        in_dim=None,
        channels=128,
        kernel_size=5,
        init_type="kaiming_normal",
        noise_scale=1.0,
        stream_sizes=(8, 20, 30),
    ):
        super().__init__()
        assert len(stream_sizes) == 3
        self.padding = (kernel_size - 1) // 2
        self.noise_scale = noise_scale
        self.stream_sizes = stream_sizes

        self.low_postfilter = _PadConv2dPostFilter(
            stream_sizes[0] + self.padding,
            channels=channels,
            kernel_size=kernel_size,
            init_type=init_type,
            padding_side="left",
        )
        self.mid_postfilter = _PadConv2dPostFilter(
            stream_sizes[1] + 2 * self.padding,
            channels=channels,
            kernel_size=kernel_size,
            init_type=init_type,
            padding_side="none",
        )
        self.high_postfilter = _PadConv2dPostFilter(
            stream_sizes[2] + self.padding,
            channels=channels,
            kernel_size=kernel_size,
            init_type=init_type,
            padding_side="right",
        )

    def forward(self, x, lengths=None, y=None):
        assert x.shape[-1] == sum(self.stream_sizes)

        # (B, T, C)
        z = torch.randn(x.shape[0], x.shape[1], 1).to(x.device) * self.noise_scale

        # Process three streams separately with a overlap width of padding
        out1 = self.low_postfilter(x[:, :, : self.stream_sizes[0] + self.padding], z)
        out2 = self.mid_postfilter(
            x[
                :,
                :,
                self.stream_sizes[0]
                - self.padding : sum(self.stream_sizes[:2])
                + self.padding,
            ],
            z,
        )
        out3 = self.high_postfilter(
            x[:, :, sum(self.stream_sizes[:2]) - self.padding :], z
        )

        # Merge the three outputs
        out = torch.cat([out1, out2, out3], dim=-1)

        return out



# -*- coding: utf-8 -*-

# Copyright 2020 Yi-Chiao Wu (Nagoya University)
# based on a Parallel WaveGAN script by Tomoki Hayashi (Nagoya University)
# (https://github.com/kan-bayashi/ParallelWaveGAN)
#  MIT License (https://opensource.org/licenses/MIT)

"""Utility functions."""

import fnmatch
import os
import sys
from logging import getLogger

import h5py
import numpy as np

# A logger for this file
logger = getLogger(__name__)


def find_files(root_dir, query="*.wav", include_root_dir=True):
    """Find files recursively.

    Args:
        root_dir (str): Root root_dir to find.
        query (str): Query to find.
        include_root_dir (bool): If False, root_dir name is not included.

    Returns:
        list: List of found filenames.

    """
    files = []
    for root, dirnames, filenames in os.walk(root_dir, followlinks=True):
        for filename in fnmatch.filter(filenames, query):
            files.append(os.path.join(root, filename))
    if not include_root_dir:
        files = [file_.replace(root_dir + "/", "") for file_ in files]

    return files


def read_hdf5(hdf5_name, hdf5_path):
    """Read hdf5 dataset.

    Args:
        hdf5_name (str): Filename of hdf5 file.
        hdf5_path (str): Dataset name in hdf5 file.

    Return:
        any: Dataset values.

    """
    if not os.path.exists(hdf5_name):
        logger.error(f"There is no such a hdf5 file ({hdf5_name}).")
        sys.exit(1)

    hdf5_file = h5py.File(hdf5_name, "r")

    if hdf5_path not in hdf5_file:
        logger.error(f"There is no such a data in hdf5 file. ({hdf5_path})")
        sys.exit(1)

    hdf5_data = hdf5_file[hdf5_path][()]
    hdf5_file.close()

    return hdf5_data


def write_hdf5(hdf5_name, hdf5_path, write_data, is_overwrite=True):
    """Write dataset to hdf5.

    Args:
        hdf5_name (str): Hdf5 dataset filename.
        hdf5_path (str): Dataset path in hdf5.
        write_data (ndarray): Data to write.
        is_overwrite (bool): Whether to overwrite dataset.

    """
    # convert to numpy array
    write_data = np.array(write_data)

    # check folder existence
    folder_name, _ = os.path.split(hdf5_name)
    if not os.path.exists(folder_name) and len(folder_name) != 0:
        os.makedirs(folder_name)

    # check hdf5 existence
    if os.path.exists(hdf5_name):
        # if already exists, open with r+ mode
        hdf5_file = h5py.File(hdf5_name, "r+")
        # check dataset existence
        if hdf5_path in hdf5_file:
            if is_overwrite:
                hdf5_file.__delitem__(hdf5_path)
            else:
                logger.error(
                    "Dataset in hdf5 file already exists. "
                    "if you want to overwrite, please set is_overwrite = True."
                )
                hdf5_file.close()
                sys.exit(1)
    else:
        # if not exists, open with w mode
        hdf5_file = h5py.File(hdf5_name, "w")

    # write data to hdf5
    hdf5_file.create_dataset(hdf5_path, data=write_data)
    hdf5_file.flush()
    hdf5_file.close()


def check_hdf5(hdf5_name, hdf5_path):
    """Check hdf5 file existence

    Args:
        hdf5_name (str): filename of hdf5 file
        hdf5_path (str): dataset name in hdf5 file

    Return:
        (bool): dataset exists then return true

    """
    if not os.path.exists(hdf5_name):
        return False
    else:
        with h5py.File(hdf5_name, "r") as f:
            if hdf5_path in f:
                return True
            else:
                return False


def read_txt(file_list):
    """Read .txt file list

    Arg:
        file_list (str): txt file filename

    Return:
        (list): list of read lines

    """
    with open(file_list, "r") as f:
        filenames = f.readlines()
    return [filename.replace("\n", "") for filename in filenames]


def check_filename(list1, list2):
    """Check the filenames of two list are matched

    Arg:
        list1 (list): file list 1
        list2 (list): file list 2

    Return:
        (bool): matched (True) or not (False)

    """

    def _filename(x):
        return os.path.basename(x).split(".")[0]

    list1 = list(map(_filename, list1))
    list2 = list(map(_filename, list2))

    return list1 == list2



from functools import partial
from pathlib import Path

import hydra
import mlflow
import torch
import torch.distributed as dist
from hydra.utils import to_absolute_path
from nnsvs.base import PredictionType
from nnsvs.mdn import mdn_get_most_probable_sigma_and_mu, mdn_loss
from nnsvs.multistream import split_streams
from nnsvs.svs import load_vocoder
from nnsvs.train_util import (
    check_resf0_config,
    collate_fn_default,
    collate_fn_random_segments,
    compute_batch_pitch_regularization_weight,
    compute_distortions,
    eval_model,
    get_stream_weight,
    log_params_from_omegaconf_dict,
    save_checkpoint,
    save_configs,
    setup,
)
from nnsvs.util import PyTorchStandardScaler, make_non_pad_mask, make_pad_mask
from omegaconf import DictConfig
from torch import nn
from torch.cuda.amp import autocast
from torch.nn.parallel import DistributedDataParallel as DDP


def train_step(
    logger,
    model,
    model_config,
    optim_config,
    optimizer,
    grad_scaler,
    train,
    in_feats,
    out_feats,
    lengths,
    out_scaler,
    feats_criterion="mse",
    pitch_reg_dyn_ws=1.0,
    pitch_reg_weight=1.0,
    stream_wise_loss=False,
    stream_weights=None,
):
    model.train() if train else model.eval()
    optimizer.zero_grad()
    log_metrics = {}

    if feats_criterion in ["l2", "mse"]:
        criterion = nn.MSELoss(reduction="none")
    elif feats_criterion in ["l1", "mae"]:
        criterion = nn.L1Loss(reduction="none")
    else:
        raise RuntimeError("not supported criterion")

    prediction_type = (
        model.module.prediction_type()
        if isinstance(model, nn.DataParallel) or isinstance(model, DDP)
        else model.prediction_type()
    )

    # Run forward
    with autocast(enabled=grad_scaler is not None):
        outs = model(in_feats, lengths, out_feats)
        if model.has_residual_lf0_prediction():
            pred_out_feats, lf0_residual = outs
        else:
            pred_out_feats, lf0_residual = outs, None

    # Mask (B, T, 1)
    with torch.no_grad():
        mask = make_non_pad_mask(lengths).unsqueeze(-1).to(in_feats.device)

    # Compute loss
    if prediction_type == PredictionType.MULTISTREAM_HYBRID:
        loss_feats = 0
        streams = split_streams(out_feats, model_config.stream_sizes)
        assert len(streams) == len(pred_out_feats)
        if stream_wise_loss:
            weights = get_stream_weight(stream_weights, model_config.stream_sizes).to(
                in_feats.device
            )
        else:
            weights = [None] * len(streams)
            N = 0

        for pred_stream, stream, sw in zip(pred_out_feats, streams, weights):
            # DDPM
            if isinstance(pred_stream, tuple) and len(pred_stream) == 2:
                noise, x_recon = pred_stream
                loss_feats_ = criterion(
                    noise.masked_select(mask), x_recon.masked_select(mask)
                )
                if stream_wise_loss:
                    loss_feats += sw * loss_feats_.mean()
                else:
                    loss_feats += loss_feats_.sum()
                    N += len(loss_feats_.view(-1))
            # MDN
            elif isinstance(pred_stream, tuple) and len(pred_stream) == 3:
                pi, sigma, mu = pred_stream
                # (B, max(T)) or (B, max(T), D_out)
                mask_ = mask if len(pi.shape) == 4 else mask.squeeze(-1)
                # Compute loss and apply mask
                with autocast(enabled=grad_scaler is not None):
                    loss_feats_ = mdn_loss(
                        pi, sigma, mu, stream, reduce=False
                    ).masked_select(mask_)
                    if stream_wise_loss:
                        loss_feats += sw * loss_feats_.mean()
                    else:
                        loss_feats += loss_feats_.sum()
                        N += len(loss_feats_.view(-1))
            else:
                # non-MDN
                with autocast(enabled=grad_scaler is not None):
                    loss_feats_ = criterion(
                        pred_stream.masked_select(mask), stream.masked_select(mask)
                    )
                    if stream_wise_loss:
                        loss_feats += sw * loss_feats_.mean()
                    else:
                        loss_feats += loss_feats_.sum()
                        N += len(loss_feats_.view(-1))
        # NOTE: Mean over batch, time and feature axis
        if not stream_wise_loss:
            loss_feats /= N
    elif prediction_type == PredictionType.PROBABILISTIC:
        pi, sigma, mu = pred_out_feats

        # (B, max(T)) or (B, max(T), D_out)
        mask_ = mask if len(pi.shape) == 4 else mask.squeeze(-1)
        # Compute loss and apply mask
        with autocast(enabled=grad_scaler is not None):
            loss_feats = mdn_loss(pi, sigma, mu, out_feats, reduce=False)
            loss_feats = loss_feats.masked_select(mask_).mean()
    else:
        with autocast(enabled=grad_scaler is not None):
            if not isinstance(pred_out_feats, list):
                # NOTE: treat as multiple predictions
                pred_out_feats = [pred_out_feats]
            loss_feats = 0
            for pred_out_feats_ in pred_out_feats:
                if stream_wise_loss:
                    weights = get_stream_weight(
                        stream_weights, model_config.stream_sizes
                    ).to(in_feats.device)
                    streams = split_streams(out_feats, model_config.stream_sizes)
                    pred_streams = split_streams(
                        pred_out_feats_, model_config.stream_sizes
                    )
                else:
                    streams = [out_feats]
                    pred_streams = [pred_out_feats_]
                    weights = [1.0]
                for pred_stream, stream, sw in zip(pred_streams, streams, weights):
                    with autocast(enabled=grad_scaler is not None):
                        loss_feats += (
                            sw
                            * criterion(
                                pred_stream.masked_select(mask),
                                stream.masked_select(mask),
                            ).mean()
                        )

    # Pitch regularization
    # NOTE: l1 loss seems to be better than mse loss in my experiments
    # we could use l2 loss as suggested in the sinsy's paper
    if pitch_reg_weight > 0.0 and lf0_residual is not None:
        with autocast(enabled=grad_scaler is not None):
            if isinstance(lf0_residual, list):
                loss_pitch = 0
                for lf0_residual_ in lf0_residual:
                    loss_pitch += (
                        (pitch_reg_dyn_ws * lf0_residual_.abs())
                        .masked_select(mask)
                        .mean()
                    )
            else:
                loss_pitch = (
                    (pitch_reg_dyn_ws * lf0_residual.abs()).masked_select(mask).mean()
                )
    else:
        loss_pitch = torch.tensor(0.0).to(in_feats.device)

    loss = loss_feats + pitch_reg_weight * loss_pitch
    if not train:
        with torch.no_grad():
            if prediction_type == PredictionType.MULTISTREAM_HYBRID:
                if len(pred_out_feats) == 4:
                    pred_mgc, pred_lf0, pred_vuv, pred_bap = pred_out_feats
                    if isinstance(pred_lf0, tuple) and len(pred_lf0) == 3:
                        pred_lf0 = mdn_get_most_probable_sigma_and_mu(*pred_lf0)[1]
                    elif isinstance(pred_lf0, tuple) and len(pred_lf0) == 2:
                        # Diffusion case: noise
                        pred_lf0 = pred_lf0[1]
                    if isinstance(pred_mgc, tuple) and len(pred_mgc) == 3:
                        pred_mgc = mdn_get_most_probable_sigma_and_mu(*pred_mgc)[1]
                    elif isinstance(pred_mgc, tuple) and len(pred_mgc) == 2:
                        # Diffusion case: noise
                        pred_mgc = pred_mgc[1]
                    if isinstance(pred_bap, tuple) and len(pred_bap) == 3:
                        pred_bap = mdn_get_most_probable_sigma_and_mu(*pred_bap)[1]
                    elif isinstance(pred_bap, tuple) and len(pred_bap) == 2:
                        # Diffusion case: noise
                        pred_bap = pred_bap[1]
                    pred_out_feats_ = torch.cat(
                        [pred_mgc, pred_lf0, pred_vuv, pred_bap], dim=-1
                    )
                elif len(pred_out_feats) == 3:
                    pred_mel, pred_lf0, pred_vuv = pred_out_feats
                    if isinstance(pred_lf0, tuple) and len(pred_lf0) == 3:
                        pred_lf0 = mdn_get_most_probable_sigma_and_mu(*pred_lf0)[1]
                    elif isinstance(pred_lf0, tuple) and len(pred_lf0) == 2:
                        # Diffusion case: noise
                        pred_lf0 = pred_lf0[1]
                    if isinstance(pred_mel, tuple) and len(pred_mel) == 3:
                        pred_mel = mdn_get_most_probable_sigma_and_mu(*pred_mel)[1]
                    elif isinstance(pred_mel, tuple) and len(pred_mel) == 2:
                        # Diffusion case: noise
                        pred_mel = pred_mel[1]
                    pred_out_feats_ = torch.cat([pred_mel, pred_lf0, pred_vuv], dim=-1)
                else:
                    raise RuntimeError("not supported")
            elif prediction_type == PredictionType.PROBABILISTIC:
                pred_out_feats_ = mdn_get_most_probable_sigma_and_mu(pi, sigma, mu)[1]
            else:
                if isinstance(pred_out_feats, list):
                    pred_out_feats_ = pred_out_feats[-1]
                else:
                    pred_out_feats_ = pred_out_feats
            distortions = compute_distortions(
                pred_out_feats_, out_feats, lengths, out_scaler, model_config
            )
            log_metrics.update(distortions)

    if train:
        if grad_scaler is not None:
            grad_scaler.scale(loss).backward()
            grad_scaler.unscale_(optimizer)
            grad_norm = torch.nn.utils.clip_grad_norm_(
                model.parameters(), optim_config.clip_norm
            )
            if not torch.isfinite(grad_norm):
                logger.info("grad norm is NaN. Skip updating")
            else:
                log_metrics["GradNorm"] = grad_norm
                grad_scaler.step(optimizer)
            grad_scaler.update()
        else:
            loss.backward()
            grad_norm = torch.nn.utils.clip_grad_norm_(
                model.parameters(), optim_config.clip_norm
            )
            if not torch.isfinite(grad_norm):
                logger.info("grad norm is NaN. Skip updating")
            else:
                log_metrics["GradNorm"] = grad_norm
                optimizer.step()

    log_metrics.update(
        {
            "Loss": loss.item(),
            "Loss_Feats": loss_feats.item(),
            "Loss_Pitch": loss_pitch.item(),
        }
    )

    return loss, log_metrics


def train_loop(
    config,
    logger,
    device,
    model,
    optimizer,
    lr_scheduler,
    grad_scaler,
    data_loaders,
    samplers,
    writer,
    in_scaler,
    out_scaler,
    use_mlflow,
    vocoder=None,
    vocoder_in_scaler=None,
    vocoder_config=None,
):
    out_dir = Path(to_absolute_path(config.train.out_dir))
    best_dev_loss = torch.finfo(torch.float32).max
    last_dev_loss = torch.finfo(torch.float32).max

    in_lf0_idx = config.data.in_lf0_idx
    in_rest_idx = config.data.in_rest_idx
    if in_lf0_idx is None or in_rest_idx is None:
        raise ValueError("in_lf0_idx and in_rest_idx must be specified")
    pitch_reg_weight = config.train.pitch_reg_weight

    if "sample_rate" not in config.data:
        logger.warning(
            "sample_rate is not found in the data config. Fallback to 48000."
        )
        sr = 48000
    else:
        sr = config.data.sample_rate

    if "feats_criterion" not in config.train:
        logger.warning(
            "feats_criterion is not found in the train config. Fallback to MSE."
        )
        feats_criterion = "mse"
    else:
        feats_criterion = config.train.feats_criterion

    if dist.is_initialized() and dist.get_rank() != 0:

        def tqdm(x, **kwargs):
            return x

    else:
        from tqdm import tqdm

    train_iter = 1
    for epoch in tqdm(range(1, config.train.nepochs + 1)):
        for phase in data_loaders.keys():
            train = phase.startswith("train")
            # https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
            if dist.is_initialized() and train and samplers[phase] is not None:
                samplers[phase].set_epoch(epoch)
            running_loss = 0
            running_metrics = {}
            evaluated = False
            for in_feats, out_feats, lengths in tqdm(
                data_loaders[phase], desc=f"{phase} iter", leave=False
            ):
                # NOTE: This is needed for pytorch's PackedSequence
                lengths, indices = torch.sort(lengths, dim=0, descending=True)
                in_feats, out_feats = (
                    in_feats[indices].to(device),
                    out_feats[indices].to(device),
                )
                # Compute denormalized log-F0 in the musical scores
                with torch.no_grad():
                    lf0_score_denorm = (
                        in_feats[:, :, in_lf0_idx] - in_scaler.min_[in_lf0_idx]
                    ) / in_scaler.scale_[in_lf0_idx]
                    # Fill zeros for rest and padded frames
                    lf0_score_denorm *= (in_feats[:, :, in_rest_idx] <= 0).float()
                    lf0_score_denorm[make_pad_mask(lengths)] = 0

                    # Compute time-variant pitch regularization weight vector
                    # NOTE: the current impl. is very slow
                    if pitch_reg_weight > 0.0:
                        pitch_reg_dyn_ws = compute_batch_pitch_regularization_weight(
                            lf0_score_denorm,
                            decay_size=config.train.pitch_reg_decay_size,
                        )
                    else:
                        pitch_reg_dyn_ws = 1.0

                if (not train) and (not evaluated):
                    eval_model(
                        phase,
                        epoch,
                        model,
                        in_feats,
                        out_feats,
                        lengths,
                        config.model,
                        out_scaler,
                        writer,
                        sr=sr,
                        lf0_score_denorm=lf0_score_denorm,
                        use_world_codec=config.data.use_world_codec,
                        vocoder=vocoder,
                        vocoder_in_scaler=vocoder_in_scaler,
                        vocoder_config=vocoder_config,
                        max_num_eval_utts=config.train.max_num_eval_utts,
                    )
                    evaluated = True

                loss, log_metrics = train_step(
                    logger=logger,
                    model=model,
                    model_config=config.model,
                    optim_config=config.train.optim,
                    optimizer=optimizer,
                    grad_scaler=grad_scaler,
                    train=train,
                    in_feats=in_feats,
                    out_feats=out_feats,
                    lengths=lengths,
                    out_scaler=out_scaler,
                    feats_criterion=feats_criterion,
                    stream_wise_loss=config.train.stream_wise_loss,
                    stream_weights=config.model.stream_weights,
                    pitch_reg_dyn_ws=pitch_reg_dyn_ws,
                    pitch_reg_weight=pitch_reg_weight,
                )

                if train:
                    if writer is not None:
                        for key, val in log_metrics.items():
                            writer.add_scalar(f"{key}_Step/{phase}", val, train_iter)
                    train_iter += 1

                running_loss += loss.item()
                for k, v in log_metrics.items():
                    try:
                        running_metrics[k] += float(v)
                    except KeyError:
                        running_metrics[k] = float(v)

            ave_loss = running_loss / len(data_loaders[phase])
            logger.info("[%s] [Epoch %s]: loss %s", phase, epoch, ave_loss)
            if writer is not None:
                writer.add_scalar(f"Loss_Epoch/{phase}", ave_loss, epoch)
            if use_mlflow:
                mlflow.log_metric(f"{phase}_loss", ave_loss, step=epoch)

            for k, v in running_metrics.items():
                ave_v = v / len(data_loaders[phase])
                if writer is not None:
                    writer.add_scalar(f"{k}_Epoch/{phase}", ave_v, epoch)
                if use_mlflow:
                    mlflow.log_metric(f"{phase}_{k}", ave_v, step=epoch)

            if not train:
                last_dev_loss = ave_loss
            if not train and ave_loss < best_dev_loss:
                best_dev_loss = ave_loss
                save_checkpoint(
                    logger, out_dir, model, optimizer, lr_scheduler, epoch, is_best=True
                )

        lr_scheduler.step()
        if epoch % config.train.checkpoint_epoch_interval == 0:
            save_checkpoint(
                logger, out_dir, model, optimizer, lr_scheduler, epoch, is_best=False
            )

    save_checkpoint(
        logger, out_dir, model, optimizer, lr_scheduler, config.train.nepochs
    )
    logger.info("The best loss was %s", best_dev_loss)
    if use_mlflow:
        mlflow.log_metric("best_dev_loss", best_dev_loss, step=epoch)
        mlflow.log_artifacts(out_dir)

    return last_dev_loss


@hydra.main(config_path="conf/train_acoustic", config_name="config")
def my_app(config: DictConfig) -> None:
    if "max_time_frames" in config.data and config.data.max_time_frames > 0:
        collate_fn = partial(
            collate_fn_random_segments, max_time_frames=config.data.max_time_frames
        )
    else:
        if "reduction_factor" in config.model.netG:
            collate_fn = partial(
                collate_fn_default,
                reduction_factor=config.model.netG.reduction_factor,
            )
        else:
            collate_fn = collate_fn_default

    if config.train.use_ddp:
        dist.init_process_group("nccl")
        rank = dist.get_rank()
        device_id = rank % torch.cuda.device_count()
        torch.cuda.set_device(device_id)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    (
        model,
        optimizer,
        lr_scheduler,
        grad_scaler,
        data_loaders,
        samplers,
        writer,
        logger,
        in_scaler,
        out_scaler,
    ) = setup(config, device, collate_fn)

    path = config.train.pretrained_vocoder_checkpoint
    if path is not None and len(path) > 0:
        logger.info(f"Loading pretrained vocoder checkpoint from {path}")
        vocoder, vocoder_in_scaler, vocoder_config = load_vocoder(
            path, device, config.model
        )
    else:
        vocoder, vocoder_in_scaler, vocoder_config = None, None, None

    check_resf0_config(logger, model, config, in_scaler, out_scaler)

    out_scaler = PyTorchStandardScaler(
        torch.from_numpy(out_scaler.mean_), torch.from_numpy(out_scaler.scale_)
    ).to(device)
    use_mlflow = config.mlflow.enabled

    if use_mlflow:
        with mlflow.start_run() as run:
            # NOTE: modify out_dir when running with mlflow
            config.train.out_dir = f"{config.train.out_dir}/{run.info.run_id}"
            save_configs(config)
            log_params_from_omegaconf_dict(config)
            last_dev_loss = train_loop(
                config=config,
                logger=logger,
                device=device,
                model=model,
                optimizer=optimizer,
                lr_scheduler=lr_scheduler,
                grad_scaler=grad_scaler,
                data_loaders=data_loaders,
                samplers=samplers,
                writer=writer,
                in_scaler=in_scaler,
                out_scaler=out_scaler,
                use_mlflow=use_mlflow,
                vocoder=vocoder,
                vocoder_in_scaler=vocoder_in_scaler,
                vocoder_config=vocoder_config,
            )
    else:
        save_configs(config)
        last_dev_loss = train_loop(
            config=config,
            logger=logger,
            device=device,
            model=model,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            grad_scaler=grad_scaler,
            data_loaders=data_loaders,
            samplers=samplers,
            writer=writer,
            in_scaler=in_scaler,
            out_scaler=out_scaler,
            use_mlflow=use_mlflow,
            vocoder=vocoder,
            vocoder_in_scaler=vocoder_in_scaler,
            vocoder_config=vocoder_config,
        )
    return last_dev_loss


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



# The code was adapted from https://github.com/jaywalnut310/vits
import torch
from torch import nn
from torch.nn import functional as F

from .attentions import MultiHeadAttention, convert_pad_shape


class LayerNorm(nn.Module):
    def __init__(self, channels, eps=1e-5):
        super().__init__()
        self.channels = channels
        self.eps = eps

        self.gamma = nn.Parameter(torch.ones(channels))
        self.beta = nn.Parameter(torch.zeros(channels))

    def forward(self, x):
        x = x.transpose(1, -1)
        x = F.layer_norm(x, (self.channels,), self.gamma, self.beta, self.eps)
        return x.transpose(1, -1)


class FFN(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        filter_channels,
        kernel_size,
        p_dropout=0.0,
        activation=None,
        causal=False,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.filter_channels = filter_channels
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.activation = activation
        self.causal = causal

        if causal:
            self.padding = self._causal_padding
        else:
            self.padding = self._same_padding

        self.conv_1 = nn.Conv1d(in_channels, filter_channels, kernel_size)
        self.conv_2 = nn.Conv1d(filter_channels, out_channels, kernel_size)
        self.drop = nn.Dropout(p_dropout)

    def forward(self, x, x_mask):
        x = self.conv_1(self.padding(x * x_mask))
        if self.activation == "gelu":
            x = x * torch.sigmoid(1.702 * x)
        else:
            x = torch.relu(x)
        x = self.drop(x)
        x = self.conv_2(self.padding(x * x_mask))
        return x * x_mask

    def _causal_padding(self, x):
        if self.kernel_size == 1:
            return x
        pad_l = self.kernel_size - 1
        pad_r = 0
        padding = [[0, 0], [0, 0], [pad_l, pad_r]]
        x = F.pad(x, convert_pad_shape(padding))
        return x

    def _same_padding(self, x):
        if self.kernel_size == 1:
            return x
        pad_l = (self.kernel_size - 1) // 2
        pad_r = self.kernel_size // 2
        padding = [[0, 0], [0, 0], [pad_l, pad_r]]
        x = F.pad(x, convert_pad_shape(padding))
        return x


class Encoder(nn.Module):
    def __init__(
        self,
        hidden_channels,
        filter_channels,
        n_heads,
        n_layers,
        kernel_size=1,
        p_dropout=0.0,
        window_size=4,
        **kwargs,
    ):
        super().__init__()
        self.hidden_channels = hidden_channels
        self.filter_channels = filter_channels
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.p_dropout = p_dropout
        self.window_size = window_size

        self.drop = nn.Dropout(p_dropout)
        self.attn_layers = nn.ModuleList()
        self.norm_layers_1 = nn.ModuleList()
        self.ffn_layers = nn.ModuleList()
        self.norm_layers_2 = nn.ModuleList()
        for _ in range(self.n_layers):
            self.attn_layers.append(
                MultiHeadAttention(
                    hidden_channels,
                    hidden_channels,
                    n_heads,
                    p_dropout=p_dropout,
                    window_size=window_size,
                )
            )
            self.norm_layers_1.append(LayerNorm(hidden_channels))
            self.ffn_layers.append(
                FFN(
                    hidden_channels,
                    hidden_channels,
                    filter_channels,
                    kernel_size,
                    p_dropout=p_dropout,
                )
            )
            self.norm_layers_2.append(LayerNorm(hidden_channels))

    def forward(self, x, x_mask):
        attn_mask = x_mask.unsqueeze(2) * x_mask.unsqueeze(-1)
        x = x * x_mask
        for i in range(self.n_layers):
            y = self.attn_layers[i](x, x, attn_mask)
            y = self.drop(y)
            x = self.norm_layers_1[i](x + y)

            y = self.ffn_layers[i](x, x_mask)
            y = self.drop(y)
            x = self.norm_layers_2[i](x + y)
        x = x * x_mask
        return x



import numpy as np
import torch
from nnsvs.base import PredictionType
from nnsvs.mdn import mdn_get_most_probable_sigma_and_mu
from torch.nn import functional as F


def predict_lf0_with_residual(
    in_feats,
    out_feats,
    in_lf0_idx=300,
    in_lf0_min=5.3936276,
    in_lf0_max=6.491111,
    out_lf0_idx=180,
    out_lf0_mean=5.953093881972361,
    out_lf0_scale=0.23435173188961034,
    residual_f0_max_cent=600,
):
    """Predict log-F0 with residual.

    Args:
        in_feats (np.ndarray): input features
        out_feats (np.ndarray): output of an acoustic model
        in_lf0_idx (int): index of LF0 in input features
        in_lf0_min (float): minimum value of LF0 in the training data of input features
        in_lf0_max (float): maximum value of LF0 in the training data of input features
        out_lf0_idx (int): index of LF0 in output features
        out_lf0_mean (float): mean of LF0 in the training data of output features
        out_lf0_scale (float): scale of LF0 in the training data of output features
        residual_f0_max_cent (int): maximum value of residual LF0 in cent

    Returns:
        tuple: (predicted log-F0, residual log-F0)
    """
    # Denormalize lf0 from input musical score
    lf0_score = in_feats[:, :, in_lf0_idx].unsqueeze(-1)
    lf0_score_denorm = lf0_score * (in_lf0_max - in_lf0_min) + in_lf0_min

    # To avoid unbounded residual f0 that would potentially cause artifacts,
    # let's constrain the residual F0 to be in a certain range by the scaled tanh
    max_lf0_ratio = residual_f0_max_cent * np.log(2) / 1200

    if len(out_feats.shape) == 4:
        # MDN case (B, T, num_gaussians, C) -> (B, T, num_gaussians)
        lf0_residual = out_feats[:, :, :, out_lf0_idx]
    else:
        # (B, T, C) -> (B, T, 1)
        lf0_residual = out_feats[:, :, out_lf0_idx].unsqueeze(-1)
    lf0_residual = max_lf0_ratio * torch.tanh(lf0_residual)

    # Residual connection in the denormalized f0 domain
    lf0_pred_denorm = lf0_score_denorm + lf0_residual

    # Back to normalized f0
    lf0_pred = (lf0_pred_denorm - out_lf0_mean) / out_lf0_scale

    return lf0_pred, lf0_residual


def pad_inference(
    model, x, lengths, reduction_factor, mode="replicate", y=None, mdn=False
):
    mod = max(lengths) % reduction_factor
    pad = reduction_factor - mod

    # Pad zeros to the end of the input features
    # so that the length of the input features is a multiple of the reduction factor
    if pad != 0:
        x_pad = F.pad(x, (0, 0, 0, pad), mode=mode)
        y_pad = F.pad(y, (0, 0, 0, pad), mode=mode) if y is not None else None
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.clone()
        else:
            lengths = lengths.copy()
        lengths = [length + pad for length in lengths]
    else:
        x_pad = x
        y_pad = y if y is not None else None
    y = model(x_pad, lengths, y_pad)
    # Residual F0 prediction: (out, lf0)
    if mdn:
        assert isinstance(y, tuple) and len(y) == 2
        # NOTE: need to parse per-stream output
        if (
            model.prediction_type() == PredictionType.MULTISTREAM_HYBRID
            and y_pad is not None
        ):
            if len(y[0]) == 4:
                mgc, lf0, vuv, bap = y[0]
                if isinstance(mgc, tuple) and len(mgc) == 3:
                    mgc = mdn_get_most_probable_sigma_and_mu(*mgc)[1]
                elif isinstance(mgc, tuple) and len(mgc) == 2:
                    mgc = mgc[1]
                if isinstance(bap, tuple) and len(bap) == 3:
                    bap = mdn_get_most_probable_sigma_and_mu(*bap)[1]
                elif isinstance(bap, tuple) and len(bap) == 2:
                    bap = bap[1]
                if pad != 0:
                    mgc = mgc[:, :-pad] if mgc.shape[1] > x.shape[1] else mgc
                    bap = bap[:, :-pad] if bap.shape[1] > x.shape[1] else bap
                    lf0 = lf0[:, :-pad] if lf0.shape[1] > x.shape[1] else lf0
                    vuv = vuv[:, :-pad] if vuv.shape[1] > x.shape[1] else vuv
                mu = torch.cat([mgc, lf0, vuv, bap], dim=-1)
            elif len(y[0]) == 3:
                mel, lf0, vuv = y[0]
                if isinstance(mel, tuple) and len(mel) == 3:
                    mel = mdn_get_most_probable_sigma_and_mu(*mel)[1]
                elif isinstance(mel, tuple) and len(mel) == 2:
                    mel = mel[1]
                if pad != 0:
                    mel = mel[:, :-pad] if mel.shape[1] > x.shape[1] else mel
                    lf0 = lf0[:, :-pad] if lf0.shape[1] > x.shape[1] else lf0
                    vuv = vuv[:, :-pad] if vuv.shape[1] > x.shape[1] else vuv
                mu = torch.cat([mel, lf0, vuv], dim=-1)
            else:
                raise ValueError("Invalid number of streams: {}".format(len(y)))
            sigma = mu
        else:
            mu, sigma = y
            if pad != 0:
                mu = mu[:, :-pad]
                sigma = sigma[:, :-pad]
        y = (mu, sigma)
    else:
        if model.has_residual_lf0_prediction():
            y = y[0]
        # Multiple output: (out, out_fine)
        if isinstance(y, list):
            y = y[-1]
        if pad != 0:
            y = y[:, :-pad]

    return y
